import logging
import codecs
import os
import numpy as np
import pandas as pd
from collections import OrderedDict
from functools import reduce
from ekorpkit.utils import print_status
from ekorpkit.utils.func import elapsed_timer
from ekorpkit import eKonf


log = logging.getLogger(__name__)


def apply_pipeline(df, pipeline, pipeline_args, update_args={}, verbose=True):
    pipeline = eKonf.to_dict(pipeline)
    pipeline_args = eKonf.to_dict(pipeline_args)

    pipeline_targets = []
    pipes = OrderedDict()
    if isinstance(pipeline, list):
        for pipe in pipeline:
            pipes[pipe] = pipe
    elif isinstance(pipeline, str):
        pipes[pipeline] = pipeline
    else:
        pipes = pipeline

    if pipes is None or len(pipes) == 0:
        log.warning("No pipeline specified")
        return df

    log.info(f"Applying pipeline: {pipes}")
    for pipe, pipe_arg_name in pipes.items():
        args = pipeline_args.get(pipe_arg_name, {}).copy()
        if pipe != pipe_arg_name:
            args_override = pipeline_args.get(pipe, {})
            args.update(args_override)
        if args and isinstance(args, dict):
            args.update(update_args)
            pipeline_targets.append(args)

    return reduce(eKonf.pipe, pipeline_targets, df)


def split_column(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    source = args.get("source")
    target = args.get("target")
    _split = args.get(eKonf.Keys.SPLIT)
    if source is None:
        log.warning("No source specified")
        return df
    log.info(f"Split column: {args}")
    df[target] = df[source].str.split(**_split)

    if verbose:
        print(df.tail())
    return df


def eval_columns(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    expressions = args.get("expressions", None)
    engine = args.get("engine", None)
    eval_at = args.get("eval_at", "dataframe")
    if expressions is None:
        log.warning("No expressions specified")
        return df
    log.info(f"Eval columns: {args}")
    if eval_at == "dataframe":
        if isinstance(expressions, list):
            for expr in expressions:
                df.eval(expr, engine=engine, inplace=True)
        else:
            for col, expr in expressions.items():
                df[col] = df.eval(expr, engine=engine)
    elif eval_at == "python":
        if isinstance(expressions, dict):
            for col, expr in expressions.items():
                df[col] = eval(expr)
        else:
            log.warning("Expressions must be a dict to eval at python")
    else:
        if isinstance(expressions, list):
            for expr in expressions:
                pd.eval(expr, engine=engine, inplace=True, target=df)
        else:
            for col, expr in expressions.items():
                df[col] = pd.eval(expr, engine=engine)

    if verbose:
        print(df.tail())
    return df


def combine_columns(df, args):
    args = eKonf.to_dict(args)
    columns = args.get("columns", None)
    if columns is None:
        log.warning("No columns specified")
        return df
    log.info(f"Combining columns: {args}")
    for col in columns:
        df[col].fillna("", inplace=True)
        df[col] = df[col].astype(str)
    separator = codecs.decode(args["separator"], "unicode_escape")
    df[args["into"]] = df[columns].agg(separator.join, axis=1)
    return df


def drop(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    labels = args.get("labels", None)
    axis = args.get("axis", 1)
    columns = args.get("columns", None)
    index = args.get("index", None)
    level = args.get("level", None)
    errors = args.get("errors", "ignore")
    if verbose:
        log.info(f"droping: {args}")
    df = df.drop(
        columns=columns,
        axis=axis,
        labels=labels,
        index=index,
        level=level,
        errors=errors,
    )
    if verbose:
        print(df.head())
    return df


def melt(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    id_vars = args.get("id_vars", None)
    value_vars = args.get("value_vars", None)
    var_name = args.get("var_name", None)
    value_name = args.get("value_name", None)
    col_level = args.get("col_level", None)
    ignore_index = args.get("ignore_index", True)
    if id_vars is None:
        if verbose:
            log.warning("No id_vars specified")
        return df
    if verbose:
        log.info(f"Melting columns: {args}")
    id_vars = list(id_vars)
    if value_vars:
        if isinstance(value_vars, str):
            value_vars = eval(value_vars)
        else:
            value_vars = list(value_vars)
    df = df.melt(
        id_vars=id_vars,
        value_vars=value_vars,
        var_name=var_name,
        value_name=value_name,
        col_level=col_level,
        ignore_index=ignore_index,
    )
    if verbose:
        print(df.head())
    return df


def plot(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    plot_cfg = args.get("visualize", {}).get("plot", {})
    if "_target_" not in plot_cfg:
        log.warning("No target specified")
        return df
    if verbose:
        log.info(f"Plotting: {plot_cfg}")
    subset = args.get("subset", {})
    col = subset.get("column", None)
    values = subset.get("values", None)
    titles = subset.get("titles", None)
    output_file = args.get("output_file", None)
    output_dir = args.get("output_dir", ".")
    if subset and col:
        title_prefix = plot_cfg["figure"].get("title", None)
        title_prefix = title_prefix + " - " if title_prefix else ""
        if isinstance(values, str):
            values = [values]
        if titles and isinstance(titles, str):
            titles = [titles]
        for i, val in enumerate(values):
            if verbose:
                log.info(f"Plotting subset: {col} == {val}")
            df_sub = df[df[col] == val]
            if output_file and "{}" in output_file:
                output_path = output_file.replace("{}", str(val))
                output_path = f"{output_dir}/{output_path}"
                plot_cfg["savefig"]["fname"] = output_path
            if titles and len(titles) > i:
                plot_cfg["figure"]["title"] = title_prefix + titles[i]
            if verbose:
                log.info(f"Plotting: {plot_cfg}")
                # print(df_sub.head())
            eKonf.instantiate(plot_cfg, df=df_sub)
    else:
        eKonf.instantiate(plot_cfg, df=df)

    return df


def pivot(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    index = args.get("index", None)
    columns = args.get("columns", None)
    values = args.get("values", None)
    fillna = args.get("fillna", None)
    reset_index = args.get("reset_index", True)
    if index is None:
        log.warning("No index specified")
        return df
    if columns is None:
        log.warning("No columns specified")
        return df
    if values is None:
        log.warning("No values specified")
        return df
    if verbose:
        log.info(f"Pivoting columns: {args}")
    if len(values) == 1:
        values = values[0]
    df = df.pivot(index=index, columns=columns, values=values)
    # df = df.set_index(index)[values].unstack()
    if fillna is not None:
        df.fillna(fillna, inplace=True)
    if reset_index:
        if verbose:
            log.info(f"Resetting index, nlevels of columns: {df.columns.nlevels}")
        df.reset_index(inplace=True)
        if df.columns.nlevels > 1:
            df.columns = ["_".join(a).strip("_") for a in df.columns.to_flat_index()]
    if verbose:
        print(df.head())
    return df


def split_sampling(df, args):
    from sklearn.model_selection import train_test_split

    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    stratify_on = args.get("stratify_on", None)
    random_state = args.get("random_state", 123)
    shuffle = args.get("shuffle", True)
    groupby = args.get("groupby", stratify_on)
    unique_key = args.get("unique_key", "id")
    test_size = args.get("test_size", 0.1)
    dev_size = args.get("dev_size", None)
    if verbose:
        log.info(f"Split sampling: {args}")

    train, dev, test = None, None, None
    if stratify_on is None:
        train, test = train_test_split(
            df, test_size=test_size, random_state=random_state, shuffle=shuffle
        )
        if dev_size:
            train, dev = train_test_split(
                train, test_size=dev_size, random_state=random_state, shuffle=shuffle
            )
    else:
        train, test = train_test_split(
            df,
            test_size=test_size,
            random_state=random_state,
            stratify=df[stratify_on],
            shuffle=shuffle,
        )
        if dev_size:
            train, dev = train_test_split(
                train,
                test_size=dev_size,
                random_state=random_state,
                stratify=train[stratify_on],
                shuffle=shuffle,
            )
    train.reset_index(drop=True, inplace=True)
    test.reset_index(drop=True, inplace=True)
    if dev_size:
        dev.reset_index(drop=True, inplace=True)
    if verbose:
        log.info(f"Total rows: {len(df)}")
        log.info(f"Train: {len(train)}")
        log.info(f"Test: {len(test)}")
        if dev_size:
            log.info(f"Dev: {len(dev)}")

        if groupby in df.columns and unique_key in df.columns:
            grp_all = (
                df.groupby(groupby)[unique_key]
                .count()
                .rename("population")
                .transform(lambda x: x / x.sum() * 100)
            )
            grp_train = (
                train.groupby(groupby)[unique_key]
                .count()
                .rename("train")
                .transform(lambda x: x / x.sum() * 100)
            )
            grp_test = (
                test.groupby(groupby)[unique_key]
                .count()
                .rename("test")
                .transform(lambda x: x / x.sum() * 100)
            )
            grp_dists = pd.concat([grp_all, grp_train, grp_test], axis=1)
            if dev_size:
                grp_dev = (
                    dev.groupby(groupby)[unique_key]
                    .count()
                    .rename("dev")
                    .transform(lambda x: x / x.sum() * 100)
                )
                grp_dists = pd.concat([grp_dists, grp_dev], axis=1)
            print(grp_dists)

    output_dir = args.get("output_dir", ".")
    train_file = args.get("train_file", None)
    test_file = args.get("test_file", None)
    dev_file = args.get("dev_file", None)
    if train_file and train is not None:
        eKonf.save_data(train, train_file, base_dir=output_dir, verbose=verbose)
    if test_file and test is not None:
        eKonf.save_data(test, test_file, base_dir=output_dir, verbose=verbose)
    if dev_file and dev is not None:
        eKonf.save_data(dev, dev_file, base_dir=output_dir, verbose=verbose)

    return df


def top_values(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    groupby = args.get("groupby", None)
    value_var = args.get("value_var", None)
    value_label = args.get("value_label", None)
    value_name = args.get("value_name", None)
    value_separator = args.get("value_separator", ", ")
    top_n = args.get("top_n", 5)
    columns = args.get("columns", None)
    use_batcher = args.get("use_batcher", True)
    minibatch_size = args.get("minibatch_size", None)

    if verbose:
        log.info(f"Split sampling: {args}")

    def label(row):
        return f"{row[value_label]}[{round(row[value_var]*100,0):.0f}%]"

    if value_label is not None and value_name is not None:
        # df[value_name] = df.apply(label, axis=1)
        df[value_name] = eKonf.apply(
            label,
            df,
            description="labeling",
            verbose=verbose,
            use_batcher=use_batcher,
            minibatch_size=minibatch_size,
        )

    top_n_grp = (
        df.sort_values(by=value_var, ascending=False)
        .groupby(groupby)
        .head(top_n)
        .reset_index(drop=True)
    )
    top_n_grp = top_n_grp.sort_values(by=groupby)
    top_n_grp = top_n_grp.groupby(groupby, as_index=False).agg(
        {value_name: value_separator.join}
    )
    top_n_grp = top_n_grp[groupby + [value_name]]
    df.drop(columns=value_name, inplace=True)
    df = pd.merge(df, top_n_grp, on=groupby, how="left")
    if columns:
        df = df[columns:]
    if verbose:
        print(df.head())

    return df


def subset(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    random_state = args.get("random_state", 123)
    head_n = args.get("head_n", None)
    tail_n = args.get("tail_n", None)
    sample_n = args.get("sample_n", None)
    sample_frac = args.get("sample_frac", None)
    if not (head_n or tail_n or sample_n or sample_frac):
        log.error("Must specify one of head_n, tail_n, sample_n, sample_frac")
        return df

    if verbose:
        log.info(f"Subsetting: {args}")
    dfs = []
    if head_n:
        dfs.append(df.head(head_n))
    if tail_n:
        dfs.append(df.tail(tail_n))
    if sample_n:
        dfs.append(df.sample(sample_n, random_state=random_state))
    if sample_frac:
        dfs.append(df.sample(frac=sample_frac, random_state=random_state))
    if verbose:
        log.info(f"Total rows: {len(df)}")

    if len(dfs) > 0:
        df = pd.concat(dfs)

    if verbose:
        log.info(f"Subset rows: {len(df)}")

    return df


def sampling(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    random_state = args.get("random_state", 123)
    groupby = args.get("groupby", None)
    sample_size_per_group = args.get("sample_size_per_group", None)
    value_var = args.get("value_var", None)
    columns = args.get("columns", None)

    if verbose:
        log.info(f"Split sampling: {args}")

    if groupby is None:
        if sample_size_per_group < 1:
            df_sample = df.sample(frac=sample_size_per_group, random_state=random_state)
        else:
            df_sample = df.sample(n=sample_size_per_group, random_state=random_state)
    else:
        if sample_size_per_group < 1:
            df_sample = df.groupby(groupby, group_keys=False).apply(
                lambda x: x.sample(
                    frac=sample_size_per_group, random_state=random_state
                )
            )
        else:
            df_sample = df.groupby(groupby, group_keys=False).apply(
                lambda x: x.sample(
                    n=min(len(x), sample_size_per_group), random_state=random_state
                )
            )
    df_sample.reset_index(drop=True, inplace=True)
    if columns:
        df_sample = df_sample[columns:]
    if verbose:
        log.info(f"Total rows: {len(df)}")
        log.info(f"Sample: {len(df_sample)}")
        print(df_sample.head())

    if groupby is not None and verbose:
        grp_all = df.groupby(groupby)[value_var].count().rename("population")
        grp_sample = df_sample.groupby(groupby)[value_var].count().rename("sample")
        grp_dists = pd.concat([grp_all, grp_sample], axis=1)
        print(grp_dists)

    save_dataframe(df_sample, args)

    return df


def aggregate_columns(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    onto_column = args.get("onto", None)
    aggregations = args.get("aggregations", None)
    reset_index = args.get("reset_index", False)
    if onto_column is None and aggregations is None:
        if verbose:
            log.warning("No columns or aggregations are specified")
        return df
    groupby_cloumns = args["groupby"]
    if groupby_cloumns is None:
        if verbose:
            log.warning("No groupby specified")
        return df
    separator = codecs.decode(args["separator"], "unicode_escape")

    num_docs = df.shape[0]
    if verbose:
        log.info(f"Aggregating columns: {args}")
    if aggregations:
        df = df.groupby(groupby_cloumns, as_index=False).agg(aggregations)
    else:
        df[onto_column].fillna("", inplace=True)
        df = df.groupby(groupby_cloumns, as_index=False).agg(
            {onto_column: separator.join}
        )
    if reset_index:
        if verbose:
            log.info(f"Resetting index, nlevels of columns: {df.columns.nlevels}")
        df.reset_index(inplace=True)
        if df.columns.nlevels > 1:
            df.columns = ["_".join(a).strip("_") for a in df.columns.to_flat_index()]
    n_docs = df.shape[0]
    if verbose:
        print(df.tail())
        log.info(f"{num_docs} documents aggregated into {n_docs} documents")
    return df


def explode_splits(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    apply_to = args.get("apply_to", "text")
    if apply_to is None:
        if verbose:
            log.warning("No columns specified")
        return df
    separator = codecs.decode(args["separator"], "unicode_escape")
    id_key = args.get("id_key", "id")
    split_key = args.get("split_key", "seg_id")
    if verbose:
        log.info(f"Exploding column: {args}")

    num_docs = df.shape[0]
    df[apply_to] = df[apply_to].str.split(separator)
    df = df.explode(apply_to)
    df[split_key] = df.groupby(id_key).cumcount()
    n_docs = df.shape[0]
    if verbose:
        log.info(f"{num_docs} documents exploded into {n_docs} documents")
    return df


def rename_columns(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    new_names = args.get("new_names", None)
    if new_names is None:
        if verbose:
            log.warning("No columns specified")
        return df
    if verbose:
        log.info(f"Renaming columns: {args}")
    if new_names is not None:
        df.rename(columns=new_names, inplace=True)
    if verbose:
        print(df.head())
    return df


def reset_index(df, args):
    verbose = args.get("verbose", False)
    index_column_name = args.get("index_column_name", None)
    if index_column_name is None:
        index_column_name = "index"
    drop_index = args.get("drop_index", False)
    if verbose:
        log.info(f"Resetting index: {args}")
    df = df.reset_index(drop=drop_index)
    if not drop_index and index_column_name != "index":
        df.rename(columns={"index": index_column_name}, inplace=True)
    if verbose:
        print(df.head())
    return df


def normalize(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    apply_to = args.get("apply_to", "text")
    if apply_to is None:
        log.warning("No columns specified")
        return df
    normalizer = args.get("preprocessor", {}).get("normalizer", None)
    if normalizer is None:
        log.warning("No normalizer specified")
        return df
    if isinstance(apply_to, str):
        apply_to = [apply_to]
    log.info("instantiating normalizer")
    normalizer = eKonf.instantiate(normalizer)
    for key in apply_to:
        with elapsed_timer(format_time=True) as elapsed:
            df[key] = eKonf.apply(
                normalizer.normalize,
                df[key],
                description=f"Normalizing column: {key}",
                verbose=verbose,
            )
            log.info(" >> elapsed time to normalize: {}".format(elapsed()))
    return df


def fillna(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    apply_to = args.get("apply_to", "text")
    if apply_to is None:
        if verbose:
            log.warning("No columns specified")
        return df
    if isinstance(apply_to, str):
        apply_to = [apply_to]
    fill_with = args.get("fill_with", "")
    if verbose:
        log.info(f"Filling missing values: {args}")
    for key in apply_to:
        df[key].fillna(fill_with, inplace=True)
    return df


def predict(df, args):
    apply_to = args.get("apply_to", "text")
    if apply_to is None:
        log.warning("No columns specified")
        return df
    if isinstance(apply_to, list):
        apply_to = apply_to[0]
    data_columns = args["data_columns"]
    if data_columns:
        df = df.copy()[data_columns]

    model = args.get(eKonf.Keys.MODEL)
    if model is None:
        log.warning("No model specified")
        return df
    model[eKonf.Keys.PREDICT][eKonf.Keys.INPUT] = apply_to
    model = eKonf.instantiate(model)

    df = model.predict(df)

    _save_data(df, args)

    return df


def aggregate_scores(df, args):
    groupby = args["groupby"]
    feature = args["feature"]
    min_examples = args["min_examples"]
    _method_ = args[eKonf.Keys.METHOD]
    if groupby is None:
        log.warning("No groupby columns specified")
        return df

    model = args[eKonf.Keys.MODEL].get("sentiment")
    if model is None:
        log.warning("No model specified")
        return df
    model = eKonf.instantiate(model)

    df = model.aggregate_scores(df, groupby, feature, min_examples, _method_=_method_)

    _save_data(df, args)

    return df


def _save_data(df, args):
    if df is None:
        log.info("No data to save")
        return
    _path = args[eKonf.Keys.PATH][eKonf.Keys.OUTPUT]
    _path[eKonf.Keys.SUFFIX.value] = args.get(eKonf.Keys.SUFFIX)
    if _path[eKonf.Keys.FILENAME]:
        log.info(f"Saving data to: {_path}")
        eKonf.save_data(df, **_path)
    else:
        log.info("filename not specified")


def segment(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    num_workers = args.get("num_workers", 1)
    use_batcher = args.get("use_batcher", True)
    minibatch_size = args.get("minibatch_size", None)
    apply_to = args.get("apply_to", "text")
    if apply_to is None:
        if verbose:
            log.warning("No columns specified")
        return df
    segmenter = args.get("preprocessor", {}).get("segmenter", None)
    if segmenter is None:
        log.warning("No segmenter specified")
        return df
    if isinstance(apply_to, str):
        apply_to = [apply_to]
    log.info("instantiating segmenter")
    segmenter = eKonf.instantiate(segmenter)
    for key in apply_to:
        with elapsed_timer(format_time=True) as elapsed:
            df[key] = eKonf.apply(
                segmenter.segment_article,
                df[key],
                description=f"Splitting column: {key}",
                verbose=verbose,
                use_batcher=use_batcher,
                minibatch_size=minibatch_size,
                num_workers=num_workers,
            )
            log.info(" >> elapsed time to segment: {}".format(elapsed()))
    return df


def tokenize(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    use_batcher = args.get("use_batcher", True)
    minibatch_size = args.get("minibatch_size", None)
    apply_to = args.get("apply_to", "text")
    if apply_to is None:
        log.warning("No columns specified")
        return df
    tokenizer = args.get("preprocessor", {}).get("tokenizer", None)
    if tokenizer is None:
        log.warning("No tokenizer specified")
        return df
    if isinstance(apply_to, str):
        apply_to = [apply_to]
    log.info("instantiating tokenizer")
    tokenizer = eKonf.instantiate(tokenizer)
    for key in apply_to:
        with elapsed_timer(format_time=True) as elapsed:
            df[key] = eKonf.apply(
                tokenizer.tokenize_article,
                df[key],
                description=f"Tokenizing column: {key}",
                verbose=verbose,
                use_batcher=use_batcher,
                minibatch_size=minibatch_size,
            )
            log.info(" >> elapsed time to segment: {}".format(elapsed()))
    return df


def extract_tokens(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    use_batcher = args.get("use_batcher", True)
    nouns_only = args.get("nouns_only", True)
    filter_stopwords_only = args.get("filter_stopwords_only", False)
    minibatch_size = args.get("minibatch_size", None)
    apply_to = args.get("apply_to", "text")
    if apply_to is None:
        log.warning("No columns specified")
        return df
    tokenizer = args.get("preprocessor", {}).get("tokenizer", None)
    if tokenizer is None:
        log.warning("No tokenizer specified")
        return df
    if isinstance(apply_to, str):
        apply_to = [apply_to]
    log.info("instantiating tokenizer")
    tokenizer = eKonf.instantiate(tokenizer)
    if filter_stopwords_only:
        extract_func = tokenizer.filter_article_stopwords
    elif nouns_only:
        extract_func = tokenizer.extract_nouns
    else:
        extract_func = tokenizer.extract_tokens
    log.info(f"extract_func: {extract_func}")

    for key in apply_to:
        with elapsed_timer(format_time=True) as elapsed:
            df[key] = eKonf.apply(
                extract_func,
                df[key],
                description=f"Extracting column: {key}",
                verbose=verbose,
                use_batcher=use_batcher,
                minibatch_size=minibatch_size,
            )
            log.info(" >> elapsed time to extract tokens: {}".format(elapsed()))
    return df


def chunk(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    use_batcher = args.get("use_batcher", True)
    minibatch_size = args.get("minibatch_size", None)
    apply_to = args.get("apply_to", "text")
    if apply_to is None:
        log.warning("No columns specified")
        return df
    segmenter = args.get("preprocessor", {}).get("segmenter", None)
    if segmenter is None:
        log.warning("No segmenter specified")
        return df
    if isinstance(apply_to, str):
        apply_to = [apply_to]
    log.info("instantiating segmenter")
    segmenter = eKonf.instantiate(segmenter)
    for key in apply_to:
        with elapsed_timer(format_time=True) as elapsed:
            df[key] = eKonf.apply(
                segmenter.chunk_article,
                df[key],
                description=f"Chunking column: {key}",
                verbose=verbose,
                use_batcher=use_batcher,
                minibatch_size=minibatch_size,
            )
            log.info(" >> elapsed time to segment: {}".format(elapsed()))
    return df


def general_function(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    apply_to = args.get("apply_to", None)
    method = args.get(eKonf.Keys.METHOD, None)
    if verbose:
        log.info(f"Running dataframe function: {args}")

    with elapsed_timer(format_time=True) as elapsed:
        if apply_to is None:
            df = getattr(df, method[eKonf.Keys.METHOD_NAME])(
                **method[eKonf.Keys.rcPARAMS]
            )
        else:
            if isinstance(apply_to, str):
                apply_to = [apply_to]
            for key in apply_to:
                log.info(f"processing column: {key}")
                df[key] = getattr(df[key], method[eKonf.Keys.METHOD_NAME])(
                    **method[eKonf.Keys.rcPARAMS]
                )

        log.info(" >> elapsed time to replace: {}".format(elapsed()))
        if verbose:
            print(df.head())
    return df


def replace_regex(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    apply_to = args.get("apply_to", "text")
    if apply_to is None:
        if verbose:
            log.warning("No columns specified")
        return df
    patterns = args.get("patterns", {})
    if patterns is None:
        if verbose:
            log.warning("No patterns specified")
        return df
    if isinstance(apply_to, str):
        apply_to = [apply_to]
    if verbose:
        log.info(f"Replacing regex: {args}")
    for key in apply_to:
        with elapsed_timer(format_time=True) as elapsed:
            for pat, repl in patterns.items():
                df[key] = df[key].str.replace(pat, repl, regex=True).str.strip()
            if verbose:
                log.info(" >> elapsed time to replace regex: {}".format(elapsed()))
    return df


def remove_startswith(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    apply_to = args.get("apply_to", "text")
    if apply_to is None:
        if verbose:
            log.warning("No columns specified")
        return df
    startswith = args.get("startswith", {})
    if startswith is None:
        if verbose:
            log.warning("No startswith text specified")
        return df
    if isinstance(apply_to, str):
        apply_to = [apply_to]
    if verbose:
        log.info(f"Remove startswith: {args}")
    for key in apply_to:
        with elapsed_timer(format_time=True) as elapsed:
            for starting_text in startswith:
                log.info(f"Remove text starting with {starting_text} from [{key}]")
                idx = df[key].str.lower().str.startswith(starting_text, na=False)
                start_pos = len(starting_text)
                df.loc[idx, key] = df.loc[idx, key].str[start_pos:].str.strip()
            if verbose:
                log.info(
                    " >> elapsed time to remove startswith: {}\n".format(elapsed())
                )
    return df


def filter_length(df, args, **kwargs):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    apply_to = args.get("apply_to", "text")
    if apply_to is None:
        log.warning("No columns specified")
        return df
    if isinstance(apply_to, str):
        apply_to = [apply_to]
    min_length = args.get("min_length", None)
    max_length = args.get("max_length", None)
    if min_length is None and max_length is None:
        log.warning("No length specified")
        return df
    add_len_column = args.get("add_len_column", False)
    len_column = args.get("len_column", "num_bytes")
    func_name = args.get("len_func", "len_bytes")
    len_func = args[eKonf.Keys.FUNC].get(func_name, None)
    len_func = eKonf.instantiate(len_func)

    df = df.copy()
    if verbose:
        log.info(f"Filtering by length: {args}")
    for key in apply_to:
        with elapsed_timer(format_time=True) as elapsed:
            _len_column = f"{key}_{len_column}"
            df[_len_column] = eKonf.apply(
                len_func, df[key], description="Calculating length"
            )
            if min_length and min_length > 0:
                n_docs = df.shape[0]
                df = df.loc[df[_len_column] >= min_length]
                log.info(
                    f"removed {(n_docs-df.shape[0])} of {n_docs} documents with length < {min_length}"
                )
            if max_length and max_length > 0:
                n_docs = df.shape[0]
                df = df.loc[df[_len_column] <= max_length]
                log.info(
                    f"removed {(n_docs-df.shape[0])} of {n_docs} documents with length > {max_length}"
                )
            log.info(" >> elapsed time to filter length: {}".format(elapsed()))
        if not add_len_column:
            df = df.drop(_len_column, axis=1)

    return df


def filter_query(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    query = args.get("query", None)
    if query is None:
        log.warning("No query specified")
        return df
    if isinstance(query, str):
        query = [query]

    if verbose:
        log.info(f"Filtering by qeury: {args}")
    with elapsed_timer(format_time=True) as elapsed:
        for qry in query:
            if verbose:
                log.info(f"Preprocessing query: {qry}")
            n_docs = df.shape[0]
            df = df.query(qry, engine="python")
            if verbose:
                log.info(f"filtered {df.shape[0]} out of {n_docs} documents by {qry}")
        if verbose:
            log.info(" >> elapsed time to filter query: {}".format(elapsed()))
    return df


def drop_duplicates(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    apply_to = args.get("apply_to", None)
    if apply_to is None:
        if verbose:
            log.warning("No columns specified")
        return df
    if isinstance(apply_to, str):
        apply_to = [apply_to]
    if verbose:
        log.info(f"Dropping duplicates: {args}")
    with elapsed_timer(format_time=True) as elapsed:
        num_docs = df.shape[0]
        df = df.drop_duplicates(subset=apply_to)
        n_docs = df.shape[0]
        if verbose:
            log.info(
                f"{n_docs} documents after dropping {(num_docs-n_docs)} duplicates from [{apply_to}]"
            )
            log.info(" >> elapsed time to drop duplicates: {}".format(elapsed()))
    return df


def save_samples(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    apply_to = args.get("apply_to", "text")

    sample_length_to_print = args.get("sample_length_to_print", 1000)
    if apply_to is None:
        if verbose:
            log.warning("No columns specified")
        return df
    if isinstance(apply_to, str):
        apply_to = [apply_to]
    num_samples_to_save = args.get("num_samples_to_save", None)
    if verbose:
        log.info(f"Saving samples: {args}")

    sample_separator = "-" * 100 + "\n"
    df_sample = df.sample(num_samples_to_save)[apply_to]
    sample_text = ""
    print_text = ""
    for i, row in df_sample.iterrows():
        for key in apply_to:
            stext = row[key]
            if len(stext) > sample_length_to_print:
                ptext = stext[:sample_length_to_print] + "..."
            else:
                ptext = stext
            sample_text += key + ": \n" + stext + "\n\n"
            print_text += key + ": \n" + ptext + "\n\n"
        sample_text += sample_separator
        print_text += sample_separator
    sample_text = sample_text.strip()
    print_text = print_text.strip()

    _path = args[eKonf.Keys.PATH][eKonf.Keys.OUTPUT]
    sample_file = _path.get("filepath") or "sample.txt"
    open(sample_file, "w", encoding="utf-8").write(sample_text)

    if verbose:
        print(sample_separator)
        print(print_text)
        log.info(f"Saved {num_samples_to_save} samples to {sample_file}")

    return df


def stdout_samples(df, args):
    from contextlib import redirect_stdout

    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    apply_to = args.get("apply_to", "text")
    # sample_length_to_print = args.get("sample_length_to_print", 1000)
    if apply_to is None:
        if verbose:
            log.warning("No columns specified")
        return df
    if isinstance(apply_to, str):
        apply_to = [apply_to]
    num_samples = args.get("num_samples", None)
    output_dir = args.get("output_dir", None)
    if output_dir is None:
        output_dir = "."
    output_file = args.get("output_file", None)
    head = args.get("head", None)
    tail = args.get("tail", None)
    if head is None:
        head = 5
    if tail is None:
        tail = 5

    if verbose:
        log.info(f"Print samples: {args}")

    sample_separator = "-" * 100 + "\n"
    df_sample = df.sample(num_samples)[apply_to]

    print_text = ""
    for i, row in df_sample.iterrows():
        for key in apply_to:
            ptext = row[key]
            print_text += key + ": \n" + ptext + "\n\n"
        print_text += sample_separator
    print_text = print_text.strip()

    if output_file is not None:
        output_file = os.path.join(output_dir, output_file)
        with open(output_file, "w", encoding="utf-8") as f:
            with redirect_stdout(f):
                print(print_text)
                print(sample_separator)
                print(df.head(head))
                print(sample_separator)
                print(df.tail(tail))
    print(print_text)
    print(sample_separator)
    print(df.head(head))
    print(sample_separator)
    print(df.tail(tail))

    if verbose:
        log.info(f"Saved {num_samples} samples to {output_file}")

    return df


def save_as_text(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    apply_to = args.get("apply_to", "text")
    corpus_name = args.get("corpus_name", eKonf.Keys.CORPUS)
    output_dir = args.get("output_dir", ".")
    output_file = args.get("output_file", None)
    doc_separator = args.get("doc_separator", "\n\n")
    if isinstance(apply_to, list):
        apply_to = apply_to[0]

    os.makedirs(os.path.abspath(output_dir), exist_ok=True)
    if output_file:
        filename = output_file
    else:
        filename = f"{corpus_name}.txt"
    output_file_path = f"{output_dir}/{filename}"

    with elapsed_timer(format_time=True) as elapsed:
        n_loaded = len(df.index)
        doc_separator = str(doc_separator).encode("utf-8").decode("unicode_escape")
        with open(output_file_path, "w") as fo:
            fo.write(doc_separator.join(df[apply_to].dropna().tolist()))
        log.info(f"Corpus is exported to {output_file_path}")
        n_sampled = len(df.index)
        status = [[" x ", corpus_name, n_loaded, n_sampled, elapsed(), filename]]

    if verbose:
        print_status(status)
    return df


def split_dataframe(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    num_splits = args.get("num_splits", 1)
    if num_splits <= 1:
        return df
    if verbose:
        log.info(f"Splitting dataframe into {num_splits} splits")
    return np.array_split(df, num_splits)


def concat_dataframes(dataframes, args):
    args = eKonf.to_dict(args)

    return eKonf.concat_data(data=dataframes, **args)


def merge_dataframe(df=None, args=None):
    if args is None:
        raise ValueError("args must be specified")
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    how = args.get("how", "inner")
    merge_on = args.get("merge_on", None)
    left_on = args.get("left_on", None)
    right_on = args.get("right_on", None)
    if merge_on is None and (right_on is None or left_on is None):
        raise ValueError("merge_on or (left_on and right_on) must be specified")
    if isinstance(merge_on, str):
        merge_on = [merge_on]
    if isinstance(left_on, str):
        left_on = [left_on]
    if isinstance(right_on, str):
        right_on = [right_on]

    df_to_merge = load_dataframe(None, args)
    if merge_on:
        df = df.merge(df_to_merge, how=how, on=merge_on)
    else:
        df = df.merge(df_to_merge, how=how, left_on=left_on, right_on=right_on)
    if verbose:
        print(df.tail())
    return df


def save_metadata(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    _path = args[eKonf.Keys.PATH][eKonf.Keys.OUTPUT]
    # filetype = args.get("filetype", None)
    column_info = args.get("column_info", None)
    split_name = args.get("split_name", None)

    if verbose:
        log.info(f"Saving metadata: {args}")

    meta_info = column_info.get("meta", None)
    meta_columns = []
    if isinstance(meta_info, dict):
        meta_columns = list(meta_info.keys())
        if eKonf.Keys.SPLIT in meta_columns and eKonf.Keys.SPLIT not in df.columns:
            df[eKonf.Keys.SPLIT] = split_name
        df_meta = df[meta_columns]
        eKonf.save_data(df_meta, **_path)

    data_info = column_info.get("data", None)
    if isinstance(data_info, dict):
        data_keys = list(data_info.keys())
        data_columns = [
            col for col in df.columns if col not in meta_columns or col in data_keys
        ]
        df = df[data_columns]

    return df


def save_dataframe(df, args):
    args = eKonf.to_dict(args)

    if df is None:
        log.warning("Dataframe is None")
        return df
    filepath = args.pop("filepath", None)
    output_dir = args.pop("output_dir", None)
    output_file = args.pop("output_file", None)

    if filepath:
        filename = filepath
        output_dir = None
    else:
        filename = output_file
    eKonf.save_data(df, filename, output_dir, **args)
    return df


def load_dataframe(df=None, args=None):
    if args is None:
        raise ValueError("args must be specified")

    args = eKonf.to_dict(args)
    filepath = args.pop("filepath", None)
    data_dir = args.pop("data_dir", None)
    data_file = args.pop("data_file", None)

    if data_file:
        filename = data_file
    else:
        filename = filepath
        data_dir = None
    return eKonf.load_data(filename, data_dir, **args)


def summary_stats(df, args):
    args = eKonf.to_dict(args)
    verbose = args.get("verbose", False)
    output_dir = args.get("output_dir", ".")
    output_file = args.get("output_file", None)
    stat_args = args.get("info", {}).get("stats", None)
    if stat_args is None:
        log.warning("No stats specified")
        return df
    if output_file is None:
        output_file = "summary_stats.csv"
    os.makedirs(os.path.abspath(output_dir), exist_ok=True)
    info_path = f"{output_dir}/{output_file}"
    stat_fn = eKonf.instantiate(stat_args)
    stats = stat_fn(df)
    eKonf.save(stats, f=info_path)
    log.info(f"Saving summary stats: {info_path}")
    if verbose:
        eKonf.pprint(stats)

    return df


def save_as_json(df, args):
    args = eKonf.to_dict(args)
    # verbose = args.get("verbose", False)
    corpus_name = args.get("corpus_name", eKonf.Keys.CORPUS)
    output_dir = args.get("output_dir", ".")
    output_file = args.get("output_file", None)
    force_ascii = args.get("force_ascii", False)

    os.makedirs(os.path.abspath(output_dir), exist_ok=True)
    if output_file:
        filename = output_file
    else:
        filename = f"{corpus_name}.json"
    output_file_path = f"{output_dir}/{filename}"

    df.to_json(output_file_path, orient="records", lines=True, force_ascii=force_ascii)
    log.info(f"Corpus is exported to {output_file_path}")
    return df


def pipeline(data=None, **cfg):
    args = eKonf.to_dict(cfg)
    # verbose = args.get("verbose", False)

    if isinstance(data, pd.DataFrame):
        df = data.copy()
    else:
        if eKonf.is_instantiatable(data):
            data = eKonf.instantiate(data)
        df = data.data.copy()

    if df is None:
        raise ValueError("No dataframe to process")

    process_pipeline = args.get(eKonf.Keys.PIPELINE, [])
    if process_pipeline is None:
        process_pipeline = []

    if len(process_pipeline) > 0:
        df = apply_pipeline(df, process_pipeline, args)
    else:
        log.warning("No pipeline specified")

    return df



import os
import requests
from bs4 import BeautifulSoup
from ekorpkit import eKonf
from ekorpkit.io.fetch.web import web_download, web_download_unzip


class EDGAR:
    def __init__(self, **args):
        self.args = eKonf.to_config(args)
        self.base_url = self.args.base_url
        self.url = self.args.url
        self.output_dir = self.args.output_dir
        os.makedirs(self.output_dir, exist_ok=True)
        self.force_download = self.args.force_download
        self.name = self.args.name

        self.build()

    def build(self):
        if self.force_download or not os.listdir(self.output_dir):
            self.download_edgar()
        else:
            print(f"{self.name} is already downloaded")

    def download_edgar(self):

        user_agent = "Mozilla/5.0"
        headers = {"User-Agent": user_agent}
        page = requests.get(self.url, headers=headers)

        soup = BeautifulSoup(page.content, "html.parser")
        filelist = soup.find_all("a", class_="filename")

        for file in filelist:
            link = self.base_url + file.get("href")
            file_path = self.output_dir + "/" + file.get_text().strip()
            web_download(link, file_path, self.name, self.force_download)



def treemap(
    df,
    fig_filepath,
    scale=1.5,
    columns=None,
    treemap=None,
    layout=None,
    update_layout=None,
    **kwargs,
):
    # textinfo = "label+value+percent parent+percent root"
    # import plotly.express as px
    import plotly.graph_objects as go

    labels = list(df[columns.label].to_list())
    values = list(df[columns.value].to_list())
    parents = list(df[columns.parent].to_list())

    layout = go.Layout(**layout)
    fig = go.Figure(
        go.Treemap(
            labels=labels,
            parents=parents,
            values=values,
            **treemap,
        ),
        layout=layout,
    )
    fig.update_layout(
        **update_layout,
    )

    fig.write_image(fig_filepath, scale=scale)



import logging
import random
from collections import Counter, defaultdict
from itertools import chain
from typing import Any, DefaultDict, Dict, List, NamedTuple, Optional, Set, Tuple, Union

import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
from munkres import Munkres  # type: ignore
from tqdm import trange

from ..analysis import LFAnalysis
from .base_labeler import BaseLabeler
from .graph_utils import get_clique_tree
from .logger import Logger
from ...types import Config
from ...utils.config_utils import merge_config
from ...utils.lr_schedulers import LRSchedulerConfig
from ...utils.optimizers import OptimizerConfig

Metrics = Dict[str, float]


class TrainConfig(Config):
    """Settings for the fit() method of LabelModel.

    Parameters
    ----------
    n_epochs
        The number of epochs to train (where each epoch is a single optimization step)
    lr
        Base learning rate (will also be affected by lr_scheduler choice and settings)
    l2
        Centered L2 regularization strength
    optimizer
        Which optimizer to use (one of ["sgd", "adam", "adamax"])
    optimizer_config
        Settings for the optimizer
    lr_scheduler
        Which lr_scheduler to use (one of ["constant", "linear", "exponential", "step"])
    lr_scheduler_config
        Settings for the LRScheduler
    prec_init
        LF precision initializations / priors
    seed
        A random seed to initialize the random number generator with
    log_freq
        Report loss every this many epochs (steps)
    mu_eps
        Restrict the learned conditional probabilities to [mu_eps, 1-mu_eps]
    """

    n_epochs: int = 100
    lr: float = 0.01
    l2: float = 0.0
    optimizer: str = "sgd"
    optimizer_config: OptimizerConfig = OptimizerConfig()  # type: ignore
    lr_scheduler: str = "constant"
    lr_scheduler_config: LRSchedulerConfig = LRSchedulerConfig()  # type: ignore
    prec_init: Union[float, List[float], np.ndarray, torch.Tensor] = 0.7
    seed: int = np.random.randint(1e6)  # type: ignore
    log_freq: int = 10
    mu_eps: Optional[float] = None


class LabelModelConfig(Config):
    """Settings for the LabelModel initialization.

    Parameters
    ----------
    verbose
        Whether to include print statements
    device
        What device to place the model on ('cpu' or 'cuda:0', for example)
    """

    verbose: bool = True
    device: str = "cpu"


class _CliqueData(NamedTuple):
    start_index: int
    end_index: int
    max_cliques: Set[int]


class LabelModel(nn.Module, BaseLabeler):
    r"""A model for learning the LF accuracies and combining their output labels.

    This class learns a model of the labeling functions' conditional probabilities
    of outputting the true (unobserved) label `Y`, `P(\lf | Y)`, and uses this learned
    model to re-weight and combine their output labels.

    This class is based on the approach in [Training Complex Models with Multi-Task
    Weak Supervision](https://arxiv.org/abs/1810.02840), published in AAAI'19. In this
    approach, we compute the inverse generalized covariance matrix of the junction tree
    of a given LF dependency graph, and perform a matrix completion-style approach with
    respect to these empirical statistics. The result is an estimate of the conditional
    LF probabilities, `P(\lf | Y)`, which are then set as the parameters of the label
    model used to re-weight and combine the labels output by the LFs.

    Currently this class uses a conditionally independent label model, in which the LFs
    are assumed to be conditionally independent given `Y`.

    Examples
    --------
    >>> label_model = LabelModel()
    >>> label_model = LabelModel(cardinality=3)
    >>> label_model = LabelModel(cardinality=3, device='cpu')
    >>> label_model = LabelModel(cardinality=3)

    Parameters
    ----------
    cardinality
        Number of classes, by default 2
    **kwargs
        Arguments for changing config defaults

    Raises
    ------
    ValueError
        If config device set to cuda but only cpu is available

    Attributes
    ----------
    cardinality
        Number of classes, by default 2
    config
        Training configuration
    seed
        Random seed
    """

    def __init__(self, cardinality: int = 2, **kwargs: Any) -> None:
        super().__init__()
        self.config: LabelModelConfig = LabelModelConfig(**kwargs)
        self.cardinality = cardinality

        # Confirm that cuda is available if config is using CUDA
        if self.config.device != "cpu" and not torch.cuda.is_available():
            raise ValueError("device=cuda but CUDA not available.")

        # By default, put model in eval mode; switch to train mode in training
        self.eval()

    def _create_L_ind(self, L: np.ndarray) -> np.ndarray:
        """Convert a label matrix with labels in 0...k to a one-hot format.

        Parameters
        ----------
        L
            An [n,m] label matrix with values in {0,1,...,k}

        Returns
        -------
        np.ndarray
            An [n,m*k] dense np.ndarray with values in {0,1}
        """
        L_ind = np.zeros((self.n, self.m * self.cardinality))
        for y in range(1, self.cardinality + 1):
            # A[x::y] slices A starting at x at intervals of y
            # e.g., np.arange(9)[0::3] == np.array([0,3,6])
            L_ind[:, (y - 1) :: self.cardinality] = np.where(L == y, 1, 0)
        return L_ind

    def _get_augmented_label_matrix(
        self, L: np.ndarray, higher_order: bool = False
    ) -> np.ndarray:
        """Create augmented version of label matrix.

        In augmented version, each column is an indicator
        for whether a certain source or clique of sources voted in a certain
        pattern.

        Parameters
        ----------
        L
            An [n,m] label matrix with values in {0,1,...,k}
        higher_order
            Whether to include higher-order correlations (e.g. LF pairs) in matrix

        Returns
        -------
        np.ndarray
            An [n,m*k] dense matrix with values in {0,1}
        """
        # Create a helper data structure which maps cliques (as tuples of member
        # sources) --> {start_index, end_index, maximal_cliques}, where
        # the last value is a set of indices in this data structure
        self.c_data: Dict[int, _CliqueData] = {}
        for i in range(self.m):
            self.c_data[i] = _CliqueData(
                start_index=i * self.cardinality,
                end_index=(i + 1) * self.cardinality,
                max_cliques=set(
                    [
                        j
                        for j in self.c_tree.nodes()
                        if i in self.c_tree.nodes[j]["members"]
                    ]
                ),
            )

        L_ind = self._create_L_ind(L)

        # Get the higher-order clique statistics based on the clique tree
        # First, iterate over the maximal cliques (nodes of c_tree) and
        # separator sets (edges of c_tree)
        if higher_order:
            L_aug = np.copy(L_ind)
            for item in chain(self.c_tree.nodes(), self.c_tree.edges()):
                if isinstance(item, int):
                    C = self.c_tree.nodes[item]
                elif isinstance(item, tuple):
                    C = self.c_tree[item[0]][item[1]]
                else:
                    raise ValueError(item)
                members = list(C["members"])

                # With unary maximal clique, just store its existing index
                C["start_index"] = members[0] * self.cardinality
                C["end_index"] = (members[0] + 1) * self.cardinality
            return L_aug
        else:
            return L_ind

    def _build_mask(self) -> None:
        """Build mask applied to O^{-1}, O for the matrix approx constraint."""
        self.mask = torch.ones(self.d, self.d).bool()
        for ci in self.c_data.values():
            si = ci.start_index
            ei = ci.end_index
            for cj in self.c_data.values():
                sj, ej = cj.start_index, cj.end_index

                # Check if ci and cj are part of the same maximal clique
                # If so, mask out their corresponding blocks in O^{-1}
                if len(ci.max_cliques.intersection(cj.max_cliques)) > 0:
                    self.mask[si:ei, sj:ej] = 0
                    self.mask[sj:ej, si:ei] = 0

    def _generate_O(self, L: np.ndarray, higher_order: bool = False) -> None:
        """Generate overlaps and conflicts matrix from label matrix.

        Parameters
        ----------
        L
            An [n,m] label matrix with values in {0,1,...,k}
        higher_order
            Whether to include higher-order correlations (e.g. LF pairs) in matrix
        """
        L_aug = self._get_augmented_label_matrix(L, higher_order=higher_order)
        self.d = L_aug.shape[1]
        self.O = (
            torch.from_numpy(L_aug.T @ L_aug / self.n).float().to(self.config.device)
        )

    def _init_params(self) -> None:
        r"""Initialize the learned params.

        - \mu is the primary learned parameter, where each row corresponds to
        the probability of a clique C emitting a specific combination of labels,
        conditioned on different values of Y (for each column); that is:

            self.mu[i*self.cardinality + j, y] = P(\lambda_i = j | Y = y)

        and similarly for higher-order cliques.

        Raises
        ------
        ValueError
            If prec_init shape does not match number of LFs
        """
        # Initialize mu so as to break basic reflective symmetry
        # Note that we are given either a single or per-LF initial precision
        # value, prec_i = P(Y=y|\lf=y), and use:
        #   mu_init = P(\lf=y|Y=y) = P(\lf=y) * prec_i / P(Y=y)

        # Handle single values
        if isinstance(self.train_config.prec_init, (int, float)):
            self._prec_init = self.train_config.prec_init * torch.ones(self.m)
        elif isinstance(self.train_config.prec_init, np.ndarray):
            self._prec_init = torch.Tensor(self.train_config.prec_init)
        elif isinstance(self.train_config.prec_init, list):
            self._prec_init = torch.Tensor(self.train_config.prec_init)
        elif not isinstance(self.train_config.prec_init, torch.Tensor):
            raise TypeError(
                f"prec_init is of type {type(self.train_config.prec_init)} which is not supported currently."
            )
        if self._prec_init.shape[0] != self.m:
            raise ValueError(f"prec_init must have shape {self.m}.")

        # Get the per-value labeling propensities
        # Note that self.O must have been computed already!
        lps = torch.diag(self.O).cpu().detach().numpy()

        # TODO: Update for higher-order cliques!
        self.mu_init = torch.zeros(self.d, self.cardinality)
        for i in range(self.m):
            for y in range(self.cardinality):
                idx = i * self.cardinality + y
                mu_init = torch.clamp(lps[idx] * self._prec_init[i] / self.p[y], 0, 1)
                self.mu_init[idx, y] += mu_init

        # Initialize randomly based on self.mu_init
        self.mu = nn.Parameter(  # type: ignore
            self.mu_init.clone() * np.random.random()
        ).float()

        # Build the mask over O^{-1}
        self._build_mask()

    def _get_conditional_probs(self, mu: np.ndarray) -> np.ndarray:
        r"""Return the estimated conditional probabilities table given parameters mu.

        Given a parameter vector mu, return the estimated conditional probabilites
        table cprobs, where cprobs is an (m, k+1, k)-dim np.ndarray with:

            cprobs[i, j, k] = P(\lf_i = j-1 | Y = k)

        where m is the number of LFs, k is the cardinality, and cprobs includes the
        conditional abstain probabilities P(\lf_i = -1 | Y = y).

        Parameters
        ----------
        mu
            An [m * k, k] np.ndarray with entries in [0, 1]

        Returns
        -------
        np.ndarray
            An [m, k + 1, k] np.ndarray conditional probabilities table.
        """
        cprobs = np.zeros((self.m, self.cardinality + 1, self.cardinality))
        for i in range(self.m):
            # si = self.c_data[(i,)]['start_index']
            # ei = self.c_data[(i,)]['end_index']
            # mu_i = mu[si:ei, :]
            mu_i = mu[i * self.cardinality : (i + 1) * self.cardinality, :]
            cprobs[i, 1:, :] = mu_i

            # The 0th row (corresponding to abstains) is the difference between
            # the sums of the other rows and one, by law of total probability
            cprobs[i, 0, :] = 1 - mu_i.sum(axis=0)
        return cprobs

    def get_conditional_probs(self) -> np.ndarray:
        r"""Return the estimated conditional probabilities table.

        Return the estimated conditional probabilites table cprobs, where cprobs is an
        (m, k+1, k)-dim np.ndarray with:

            cprobs[i, j, k] = P(\lf_i = j-1 | Y = k)

        where m is the number of LFs, k is the cardinality, and cprobs includes the
        conditional abstain probabilities P(\lf_i = -1 | Y = y).

        Returns
        -------
        np.ndarray
            An [m, k + 1, k] np.ndarray conditional probabilities table.
        """
        return self._get_conditional_probs(self.mu.cpu().detach().numpy())

    def get_weights(self) -> np.ndarray:
        """Return the vector of learned LF weights for combining LFs.

        Returns
        -------
        np.ndarray
            [m,1] vector of learned LF weights for combining LFs.

        Example
        -------
        >>> L = np.array([[1, 1, 1], [1, 1, -1], [-1, 0, 0], [0, 0, 0]])
        >>> label_model = LabelModel(verbose=False)
        >>> label_model.fit(L, seed=123)
        >>> np.around(label_model.get_weights(), 2)  # doctest: +SKIP
        array([0.99, 0.99, 0.99])
        """
        accs = np.zeros(self.m)
        cprobs = self.get_conditional_probs()
        for i in range(self.m):
            accs[i] = np.diag(cprobs[i, 1:, :] @ self.P.cpu().detach().numpy()).sum()
        return np.clip(accs / self.coverage, 1e-6, 1.0)

    def predict_proba(self, L: np.ndarray) -> np.ndarray:
        r"""Return label probabilities P(Y | \lambda).

        Parameters
        ----------
        L
            An [n,m] matrix with values in {-1,0,1,...,k-1}f

        Returns
        -------
        np.ndarray
            An [n,k] array of probabilistic labels

        Example
        -------
        >>> L = np.array([[0, 0, 0], [1, 1, 1], [1, 1, 1]])
        >>> label_model = LabelModel(verbose=False)
        >>> label_model.fit(L, seed=123)
        >>> np.around(label_model.predict_proba(L), 1)  # doctest: +SKIP
        array([[1., 0.],
               [0., 1.],
               [0., 1.]])
        """
        L_shift = L + 1  # convert to {0, 1, ..., k}
        self._set_constants(L_shift)
        L_aug = self._get_augmented_label_matrix(L_shift)
        mu = self.mu.cpu().detach().numpy()
        jtm = np.ones(L_aug.shape[1])

        # Note: We omit abstains, effectively assuming uniform distribution here
        X = np.exp(L_aug @ np.diag(jtm) @ np.log(mu) + np.log(self.p))
        Z = np.tile(X.sum(axis=1).reshape(-1, 1), self.cardinality)
        return X / Z

    def predict(
        self,
        L: np.ndarray,
        return_probs: Optional[bool] = False,
        tie_break_policy: str = "abstain",
    ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """Return predicted labels, with ties broken according to policy.

        Policies to break ties include:

        - "abstain": return an abstain vote (-1)
        - "true-random": randomly choose among the tied options
        - "random": randomly choose among tied option using deterministic hash

        NOTE: if tie_break_policy="true-random", repeated runs may have slightly different
        results due to difference in broken ties


        Parameters
        ----------
        L
            An [n,m] matrix with values in {-1,0,1,...,k-1}
        return_probs
            Whether to return probs along with preds
        tie_break_policy
            Policy to break ties when converting probabilistic labels to predictions

        Returns
        -------
        np.ndarray
            An [n,1] array of integer labels

        (np.ndarray, np.ndarray)
            An [n,1] array of integer labels and an [n,k] array of probabilistic labels


        Example
        -------
        >>> L = np.array([[0, 0, -1], [1, 1, -1], [0, 0, -1]])
        >>> label_model = LabelModel(verbose=False)
        >>> label_model.fit(L)
        >>> label_model.predict(L)
        array([0, 1, 0])
        """
        return super(LabelModel, self).predict(L, return_probs, tie_break_policy)

    def score(
        self,
        L: np.ndarray,
        Y: np.ndarray,
        metrics: Optional[List[str]] = ["accuracy"],
        tie_break_policy: str = "abstain",
    ) -> Dict[str, float]:
        """Calculate one or more scores from user-specified and/or user-defined metrics.

        Parameters
        ----------
        L
            An [n,m] matrix with values in {-1,0,1,...,k-1}
        Y
            Gold labels associated with data points in L
        metrics
            A list of metric names. Possbile metrics are - `accuracy`, `coverage`,
            `precision`, `recall`, `f1`, `f1_micro`, `f1_macro`, `fbeta`,
            `matthews_corrcoef`, `roc_auc`. See `sklearn.metrics
            <https://scikit-learn.org/stable/modules/classes.html#module-sklearn.metrics>`_
            for details on the metrics.
        tie_break_policy
            Policy to break ties when converting probabilistic labels to predictions.
            Same as :func:`.predict` method above.


        Returns
        -------
        Dict[str, float]
            A dictionary mapping metric names to metric scores

        Example
        -------
        >>> L = np.array([[1, 1, -1], [0, 0, -1], [1, 1, -1]])
        >>> label_model = LabelModel(verbose=False)
        >>> label_model.fit(L)
        >>> label_model.score(L, Y=np.array([1, 1, 1]))
        {'accuracy': 0.6666666666666666}
        >>> label_model.score(L, Y=np.array([1, 1, 1]), metrics=["f1"])
        {'f1': 0.8}
        """
        return super(LabelModel, self).score(L, Y, metrics, tie_break_policy)

    # These loss functions get all their data directly from the LabelModel
    # (for better or worse). The unused *args make these compatible with the
    # Classifer._train() method which expect loss functions to accept an input.
    def _loss_l2(self, l2: float = 0) -> torch.Tensor:
        r"""L2 loss centered around mu_init, scaled optionally per-source.

        In other words, diagonal Tikhonov regularization,
            ||D(\mu-\mu_{init})||_2^2
        where D is diagonal.

        Parameters
        ----------
        l2
            A float or np.array representing the per-source regularization
            strengths to use, by default 0

        Returns
        -------
        torch.Tensor
            L2 loss between learned mu and initial mu
        """
        if isinstance(l2, (int, float)):
            D = l2 * torch.eye(self.d)
        else:
            D = torch.diag(torch.from_numpy(l2)).type(torch.float32)
        D = D.to(self.config.device)
        # Note that mu is a matrix and this is the *Frobenius norm*
        return torch.norm(D @ (self.mu - self.mu_init)) ** 2

    def _loss_mu(self, l2: float = 0) -> torch.Tensor:
        r"""Overall mu loss.

        Parameters
        ----------
        l2
            A float or np.array representing the per-source regularization
                strengths to use, by default 0

        Returns
        -------
        torch.Tensor
            Overall mu loss between learned mu and initial mu
        """
        loss_1 = torch.norm((self.O - self.mu @ self.P @ self.mu.t())[self.mask]) ** 2
        loss_2 = torch.norm(torch.sum(self.mu @ self.P, 1) - torch.diag(self.O)) ** 2
        return loss_1 + loss_2 + self._loss_l2(l2=l2)

    def _set_class_balance(
        self,
        class_balance: Optional[List[float]],
        Y_dev: Optional[np.ndarray] = None,
    ) -> None:
        """Set a prior for the class balance.

        In order of preference:
        1) Use user-provided class_balance
        2) Estimate balance from Y_dev
        3) Assume uniform class distribution
        """
        if class_balance is not None:
            self.p = np.array(class_balance)
            if len(self.p) != self.cardinality:
                raise ValueError(
                    f"class_balance has {len(self.p)} entries. Does not match LabelModel cardinality {self.cardinality}."
                )
        elif Y_dev is not None:
            class_counts = Counter(Y_dev)
            sorted_counts = np.array([v for k, v in sorted(class_counts.items())])
            self.p = sorted_counts / sum(sorted_counts)
            if len(self.p) != self.cardinality:
                raise ValueError(
                    f"Y_dev has {len(self.p)} class(es). Does not match LabelModel cardinality {self.cardinality}."
                )
        else:
            self.p = (1 / self.cardinality) * np.ones(self.cardinality)

        if np.any(self.p == 0):
            raise ValueError(
                f"Class balance prior is 0 for class(es) {np.where(self.p)[0]}."
            )
        self.P = torch.diag(torch.from_numpy(self.p)).float().to(self.config.device)

    def _set_constants(self, L: np.ndarray) -> None:
        self.n, self.m = L.shape
        if self.m < 3:
            raise ValueError("L_train should have at least 3 labeling functions")
        self.t = 1

    def _create_tree(self) -> None:
        nodes = range(self.m)
        self.c_tree = get_clique_tree(nodes, [])

    def _execute_logging(self, loss: torch.Tensor) -> Metrics:
        self.eval()
        self.running_examples: int
        self.running_loss: float
        self.running_loss += loss.item()
        self.running_examples += 1

        # Always add average loss
        metrics_dict = {"train/loss": self.running_loss / self.running_examples}

        if self.logger.check():
            if self.config.verbose:
                self.logger.log(metrics_dict)

            # Reset running loss and examples counts
            self.running_loss = 0.0
            self.running_examples = 0

        self.train()
        return metrics_dict

    def _set_logger(self) -> None:
        self.logger = Logger(self.train_config.log_freq)
        if self.config.verbose:
            logging.basicConfig(level=logging.INFO)

    def _set_optimizer(self) -> None:
        parameters = filter(lambda p: p.requires_grad, self.parameters())

        optimizer_config = self.train_config.optimizer_config
        optimizer_name = self.train_config.optimizer
        optimizer: optim.Optimizer  # type: ignore

        if optimizer_name == "sgd":
            optimizer = optim.SGD(  # type: ignore
                parameters,
                lr=self.train_config.lr,
                weight_decay=self.train_config.l2,
                **optimizer_config.sgd_config._asdict(),
            )
        elif optimizer_name == "adam":
            optimizer = optim.Adam(
                parameters,
                lr=self.train_config.lr,
                weight_decay=self.train_config.l2,
                **optimizer_config.adam_config._asdict(),
            )
        elif optimizer_name == "adamax":
            optimizer = optim.Adamax(  # type: ignore
                parameters,
                lr=self.train_config.lr,
                weight_decay=self.train_config.l2,
                **optimizer_config.adamax_config._asdict(),
            )
        else:
            raise ValueError(f"Unrecognized optimizer option '{optimizer_name}'")

        self.optimizer = optimizer

    def _set_lr_scheduler(self) -> None:
        # Set warmup scheduler
        self._set_warmup_scheduler()

        # Set lr scheduler
        lr_scheduler_name = self.train_config.lr_scheduler
        lr_scheduler_config = self.train_config.lr_scheduler_config
        lr_scheduler: Optional[optim.lr_scheduler._LRScheduler]

        if lr_scheduler_name == "constant":
            lr_scheduler = None
        elif lr_scheduler_name == "linear":
            total_steps = self.train_config.n_epochs
            linear_decay_func = lambda x: (total_steps - self.warmup_steps - x) / (
                total_steps - self.warmup_steps
            )
            lr_scheduler = optim.lr_scheduler.LambdaLR(  # type: ignore
                self.optimizer, linear_decay_func
            )
        elif lr_scheduler_name == "exponential":
            lr_scheduler = optim.lr_scheduler.ExponentialLR(
                self.optimizer, **lr_scheduler_config.exponential_config._asdict()
            )
        elif lr_scheduler_name == "step":
            lr_scheduler = optim.lr_scheduler.StepLR(
                self.optimizer, **lr_scheduler_config.step_config._asdict()
            )
        else:
            raise ValueError(f"Unrecognized lr scheduler option '{lr_scheduler_name}'")

        self.lr_scheduler = lr_scheduler

    def _set_warmup_scheduler(self) -> None:
        warmup_scheduler: Optional[optim.lr_scheduler.LambdaLR]

        if self.train_config.lr_scheduler_config.warmup_steps:
            warmup_steps = self.train_config.lr_scheduler_config.warmup_steps
            if warmup_steps < 0:
                raise ValueError("warmup_steps much greater or equal than 0.")
            warmup_unit = self.train_config.lr_scheduler_config.warmup_unit
            if warmup_unit == "epochs":
                self.warmup_steps = int(warmup_steps)
            else:
                raise ValueError(
                    "LabelModel does not support any warmup_unit other than 'epochs'."
                )

            linear_warmup_func = lambda x: x / self.warmup_steps
            warmup_scheduler = optim.lr_scheduler.LambdaLR(  # type: ignore
                self.optimizer, linear_warmup_func
            )
            if self.config.verbose:  # pragma: no cover
                logging.info(f"Warmup {self.warmup_steps} steps.")

        elif self.train_config.lr_scheduler_config.warmup_percentage:
            warmup_percentage = self.train_config.lr_scheduler_config.warmup_percentage
            self.warmup_steps = int(warmup_percentage * self.train_config.n_epochs)
            linear_warmup_func = lambda x: x / self.warmup_steps
            warmup_scheduler = optim.lr_scheduler.LambdaLR(  # type: ignore
                self.optimizer, linear_warmup_func
            )
            if self.config.verbose:  # pragma: no cover
                logging.info(f"Warmup {self.warmup_steps} steps.")

        else:
            warmup_scheduler = None
            self.warmup_steps = 0

        self.warmup_scheduler = warmup_scheduler

    def _update_lr_scheduler(self, step: int) -> None:
        if self.warmup_scheduler and step < self.warmup_steps:
            self.warmup_scheduler.step()  # type: ignore
        elif self.lr_scheduler is not None:
            self.lr_scheduler.step()  # type: ignore
            min_lr = self.train_config.lr_scheduler_config.min_lr
            if min_lr and self.optimizer.param_groups[0]["lr"] < min_lr:
                self.optimizer.param_groups[0]["lr"] = min_lr

    def _clamp_params(self) -> None:
        """Clamp the values of the learned parameter vector.

        Clamp the entries of self.mu to be in [mu_eps, 1 - mu_eps], where mu_eps is
        either set by the user, or defaults to 1 / 10 ** np.ceil(np.log10(self.n)).

        Note that if mu_eps is set too high, e.g. in sparse settings where LFs
        mostly abstain, this will result in learning conditional probabilities all
        equal to mu_eps (and/or 1 - mu_eps)!  See issue #1422.

        Note: Use user-provided value of mu_eps in train_config, else default to
            mu_eps = 1 / 10 ** np.ceil(np.log10(self.n))
        this rounding is done to make it more obvious when the parameters have been
        clamped.
        """
        if self.train_config.mu_eps is not None:
            mu_eps = self.train_config.mu_eps
        else:
            mu_eps = min(0.01, 1 / 10 ** np.ceil(np.log10(self.n)))
        self.mu.data = self.mu.clamp(mu_eps, 1 - mu_eps)  # type: ignore

    def _break_col_permutation_symmetry(self) -> None:
        r"""Heuristically choose amongst (possibly) several valid mu values.

        If there are several values of mu that equivalently satisfy the optimization
        objective, as there often are due to column permutation symmetries, then pick
        the solution that trusts the user-written LFs most.

        In more detail, suppose that mu satisfies (minimizes) the two loss objectives:
            1. O = mu @ P @ mu.T
            2. diag(O) = sum(mu @ P, axis=1)
        Then any column permutation matrix Z that commutes with P will also equivalently
        satisfy these objectives, and thus is an equally valid (symmetric) solution.
        Therefore, we select the solution that maximizes the summed probability of the
        LFs being accurate when not abstaining.

            \sum_lf \sum_{y=1}^{cardinality} P(\lf = y, Y = y)
        """
        mu = self.mu.cpu().detach().numpy()
        P = self.P.cpu().detach().numpy()
        d, k = mu.shape
        # We want to maximize the sum of diagonals of matrices for each LF. So
        # we start by computing the sum of conditional probabilities here.
        probs_sum = sum([mu[i : i + k] for i in range(0, self.m * k, k)]) @ P

        munkres_solver = Munkres()
        Z = np.zeros([k, k])

        # Compute groups of indicess with equal prior in P.
        groups: DefaultDict[float, List[int]] = defaultdict(list)
        for i, f in enumerate(P.diagonal()):
            groups[np.around(f, 3)].append(i)
        for group in groups.values():
            if len(group) == 1:
                Z[group[0], group[0]] = 1.0  # Identity permutation
                continue
            # Compute submatrix corresponding to the group.
            probs_proj = probs_sum[[[g] for g in group], group]
            # Use the Munkres algorithm to find the optimal permutation.
            # We use minus because we want to maximize diagonal sum, not minimize,
            # and transpose because we want to permute columns, not rows.
            permutation_pairs = munkres_solver.compute(-probs_proj.T)
            for i, j in permutation_pairs:
                Z[group[i], group[j]] = 1.0

        # Set mu according to permutation
        self.mu = nn.Parameter(  # type: ignore
            torch.Tensor(mu @ Z).to(self.config.device)  # type: ignore
        )

    def fit(
        self,
        L_train: np.ndarray,
        Y_dev: Optional[np.ndarray] = None,
        class_balance: Optional[List[float]] = None,
        progress_bar: bool = True,
        **kwargs: Any,
    ) -> None:
        """Train label model.

        Train label model to estimate mu, the parameters used to combine LFs.

        Parameters
        ----------
        L_train
            An [n,m] matrix with values in {-1,0,1,...,k-1}
        Y_dev
            Gold labels for dev set for estimating class_balance, by default None
        class_balance
            Each class's percentage of the population, by default None
        progress_bar
            To display a progress bar, by default True
        **kwargs
            Arguments for changing train config defaults.

            n_epochs
                The number of epochs to train (where each epoch is a single
                optimization step), default is 100
            lr
                Base learning rate (will also be affected by lr_scheduler choice
                and settings), default is 0.01
            l2
                Centered L2 regularization strength, default is 0.0
            optimizer
                Which optimizer to use (one of ["sgd", "adam", "adamax"]),
                default is "sgd"
            optimizer_config
                Settings for the optimizer
            lr_scheduler
                Which lr_scheduler to use (one of ["constant", "linear",
                "exponential", "step"]), default is "constant"
            lr_scheduler_config
                Settings for the LRScheduler
            prec_init
                LF precision initializations / priors, default is 0.7
            seed
                A random seed to initialize the random number generator with
            log_freq
                Report loss every this many epochs (steps), default is 10
            mu_eps
                Restrict the learned conditional probabilities to
                [mu_eps, 1-mu_eps], default is None

        Raises
        ------
        Exception
            If loss in NaN

        Examples
        --------
        >>> L = np.array([[0, 0, -1], [-1, 0, 1], [1, -1, 0]])
        >>> Y_dev = [0, 1, 0]
        >>> label_model = LabelModel(verbose=False)
        >>> label_model.fit(L)
        >>> label_model.fit(L, Y_dev=Y_dev, seed=2020, lr=0.05)
        >>> label_model.fit(L, class_balance=[0.7, 0.3], n_epochs=200, l2=0.4)
        """
        # Set random seed
        self.train_config: TrainConfig = merge_config(  # type:ignore
            TrainConfig(), kwargs  # type:ignore
        )
        # Update base config so that it includes all parameters
        random.seed(self.train_config.seed)
        np.random.seed(self.train_config.seed)
        torch.manual_seed(self.train_config.seed)

        # Set Logger
        self._set_logger()

        L_shift = L_train + 1  # convert to {0, 1, ..., k}
        if L_shift.max() > self.cardinality:
            raise ValueError(
                f"L_train has cardinality {L_shift.max()}, cardinality={self.cardinality} passed in."
            )

        self._set_constants(L_shift)
        self._set_class_balance(class_balance, Y_dev)
        self._create_tree()
        lf_analysis = LFAnalysis(L_train)
        self.coverage = lf_analysis.lf_coverages()

        # Compute O and initialize params
        if self.config.verbose:  # pragma: no cover
            logging.info("Computing O...")
        self._generate_O(L_shift)
        self._init_params()

        # Estimate \mu
        if self.config.verbose:  # pragma: no cover
            logging.info(r"Estimating \mu...")

        # Set model to train mode
        self.train()

        # Move model to GPU
        self.mu_init = self.mu_init.to(self.config.device)
        if self.config.verbose and self.config.device != "cpu":  # pragma: no cover
            logging.info("Using GPU...")
        self.to(self.config.device)

        # Set training components
        self._set_optimizer()
        self._set_lr_scheduler()

        # Restore model if necessary
        start_iteration = 0

        # Train the model
        metrics_hist = {}  # The most recently seen value for all metrics

        if progress_bar:
            epochs = trange(start_iteration, self.train_config.n_epochs, unit="epoch")
        else:
            epochs = range(start_iteration, self.train_config.n_epochs)

        for epoch in epochs:
            self.running_loss = 0.0
            self.running_examples = 0

            # Zero the parameter gradients
            self.optimizer.zero_grad()

            # Forward pass to calculate the average loss per example
            loss = self._loss_mu(l2=self.train_config.l2)
            if torch.isnan(loss):
                msg = "Loss is NaN. Consider reducing learning rate."
                raise Exception(msg)

            # Backward pass to calculate gradients
            # Loss is an average loss per example
            loss.backward()

            # Perform optimizer step
            self.optimizer.step()

            # Calculate metrics, log, and checkpoint as necessary
            metrics_dict = self._execute_logging(loss)
            metrics_hist.update(metrics_dict)

            # Update learning rate
            self._update_lr_scheduler(epoch)

        # Cleanup progress bar if enabled
        if progress_bar:
            epochs.close()

        # Post-processing operations on mu
        self._clamp_params()
        self._break_col_permutation_symmetry()

        # Return model to eval mode
        self.eval()

        # Print confusion matrix if applicable
        if self.config.verbose:  # pragma: no cover
            logging.info("Finished Training")



import os
import codecs
import pandas as pd
from pathlib import Path
from glob import glob
from ekorpkit.utils.func import ordinal


class ESGReport:
    def __init__(self, name, output_dir, output_file, input_path, txt_info, **kwargs):

        self.name = name
        self.input_path = input_path
        self.txt_info = txt_info
        self.output_path = f"{output_dir}/{output_file}"
        os.makedirs(output_dir, exist_ok=True)

        self.parse_text_files()

    def parse_text_files(self):
        filepaths = glob(self.input_path, recursive=True)
        filepaths = [fp for fp in filepaths if Path(fp).is_file()]
        txt_info = self.txt_info

        initial_file_num = txt_info.get("initial_file_num", 0)
        segment_separator = txt_info.get("segment_separator", None)
        segment_separator = codecs.decode(segment_separator, "unicode_escape")
        doc_id_format = txt_info.get("doc_id_format", None)
        file_num_increment = txt_info.get("file_num_increment", 1)

        file_num = initial_file_num
        reports = []
        for i, file in enumerate(filepaths):
            file = Path(file)
            print(" >> processing {} file: {}".format(ordinal(i + 1), file.name))

            texts = file.open().read()
            for seg_num, doc in enumerate(texts.split(segment_separator)):
                doc = doc.strip()
                if len(doc) > 0:
                    doc_id = doc_id_format.format(file_num=file_num, seg_num=seg_num)
                    rpt = {"doc_id": doc_id, "filename": file.stem, "text": doc}
                    reports.append(rpt)
            file_num += file_num_increment

        df = pd.DataFrame(reports)
        print(df.tail())
        df.to_csv(self.output_path, header=True, index=False)
        print(
            f"Corpus [{self.name}] is built to [{self.output_path}] from [{self.input_path}]"
        )



import logging
import codecs
import re
import math
from abc import ABCMeta, abstractmethod
from collections import namedtuple
from typing import Callable
from ekorpkit import eKonf


log = logging.getLogger(__name__)


Offset = namedtuple("Offset", ["this", "next", "prev_cum_len", "this_cum_len"])


class BaseSegmenter:
    __metaclass__ = ABCMeta
    """Abstract segmenter class from which all segmenters inherit.
    Subclasses must implement a ``segment()`` method.
    """

    @abstractmethod
    def segment(self, text):
        raise NotImplementedError("Must override segment")

    def __call__(self, text):
        """Calling a segmenter instance like a function just calls the segment method."""
        return self.segment(text)


class Segmenter(BaseSegmenter):
    def __init__(
        self,
        separators={},
        merge={},
        split={},
        filter_language={},
        filter_programming_language=False,
        filter_sentence_length={},
        chunk={},
        return_as_list=True,
        verbose=False,
        print_args=False,
        **kwargs,
    ):
        _separators = separators or {}
        _merge = merge or {}
        _split = split or {}
        _filter_language = filter_language or {}
        _filter_sentence_length = filter_sentence_length or {}
        _chunk = chunk or {}

        in_segment_separator = _separators.get("in_segment", "\n\n")
        in_sentence_separator = _separators.get("in_sentence", "\n")
        out_segment_separator = _separators.get("out_segment", "\n\n")
        out_sentence_separator = _separators.get("out_sentence", "\n")

        keep_segment = _split.get("keep_segment", True)
        max_split_length = _split.get("max_split_length", 30_000)
        max_split_iterations = _split.get("max_split_iterations", 100)

        merge_lines = _merge.get("merge_lines", False)
        merge_level = _merge.get("merge_level", "segmenmt")
        empty_lines_threshold = _merge.get("empty_lines_threshold", 0.6)
        broken_lines_threshold = _merge.get("broken_lines_threshold", 0.4)

        filter_language = _filter_language.get("filter", False)
        detection_level = _filter_language.get("detection_level", "segment")
        languages_to_keep = _filter_language.get("languages_to_keep", ["en", "ko"])
        min_language_probability = _filter_language.get("min_language_probability", 0.8)

        filter_sentence_length = _filter_sentence_length.get("filter", False)
        min_num_words = _filter_sentence_length.get("min_num_words", 3)
        min_length = _filter_sentence_length.get("min_length", 10)

        chunk_size = _chunk.get("chunk_size", 300)
        chunk_overlap = _chunk.get("chunk_overlap", False)
        self.len_func_name = _chunk.get("len_func", "len_bytes")
        len_func = _chunk.get(eKonf.Keys.FUNC, {}).get(self.len_func_name, None)
        self.len_func = eKonf.partial(len_func)

        self._in_segment_separator = codecs.decode(
            in_segment_separator, "unicode_escape"
        )
        self._in_sentence_separator = codecs.decode(
            in_sentence_separator, "unicode_escape"
        )
        self._out_segment_separator = codecs.decode(
            out_segment_separator, "unicode_escape"
        )
        self._out_sentence_separator = codecs.decode(
            out_sentence_separator, "unicode_escape"
        )
        self._return_as_list = return_as_list
        self._max_split_length = max_split_length
        self._max_split_iterations = max_split_iterations
        self._keep_segment = keep_segment
        self._merge_lines = (
            merge_lines
            if isinstance(merge_lines, bool)
            else str(merge_lines).lower() in ["true", "t", "1"]
        )
        self._merge_level = merge_level
        self._empty_lines_threshold = empty_lines_threshold
        self._broken_lines_threshold = broken_lines_threshold
        self._filter_language = filter_language
        self._detection_level = detection_level
        self._languages_to_keep = list(languages_to_keep)
        self._min_language_probability = min_language_probability
        self._filter_programming_language = filter_programming_language
        self._filter_sentence_length = filter_sentence_length
        self._min_num_words = min_num_words
        self._min_length = min_length
        self._chunk_size = chunk_size
        self._chunk_overlap = chunk_overlap
        self.verbose = verbose
        self._kwargs = kwargs
        if print_args:
            log.info(f"in_segment_separator = {self._in_segment_separator}")
            log.info(f"in_sentence_separator = {self._in_sentence_separator}")
            log.info(f"out_segment_separator = {self._out_segment_separator}")
            log.info(f"out_sentence_separator = {self._out_sentence_separator}")
            log.info(f"return_as_list = {self._return_as_list}")
            log.info(f"max_split_length = {self._max_split_length}")
            log.info(f"max_split_iterations = {self._max_split_iterations}")
            log.info(f"keep_segment = {self._keep_segment}")
            log.info(f"merge_lines = {self._merge_lines}")
            log.info(f"merge_level = {self._merge_level}")
            log.info(f"empty_lines_threshold = {self._empty_lines_threshold}")
            log.info(f"broken_lines_threshold = {self._broken_lines_threshold}")
            log.info(f"filter_language = {self._filter_language}")
            log.info(f"detection_level = {self._detection_level}")
            log.info(f"languages_to_keep = {self._languages_to_keep}")
            log.info(f"min_language_probability = {self._min_language_probability}")
            log.info(
                f"filter_programming_language = {self._filter_programming_language}"
            )
            log.info(f"filter_sentence_length = {self._filter_sentence_length}")
            log.info(f"min_num_words = {self._min_num_words}")
            log.info(f"min_length = {self._min_length}")
            log.info(f"chunk_size = {self._chunk_size}")
            log.info(f"chunk_overlap = {self._chunk_overlap}")
            log.info(f"verbose = {self.verbose}")
            log.info(f"kwargs = {self._kwargs}")

        self._guess = None

    def _check_if_merge_lines(self, text):
        num_empty_lines = 0
        num_lines_endswith_period = 0
        num_lines = len(text.split("\n"))
        for line in text.split("\n"):
            if len(line.strip()) == 0:
                num_empty_lines += 1
            if line.strip().endswith("."):
                num_lines_endswith_period += 1
        empty_lines_ratio = num_empty_lines / num_lines if num_lines > 0 else 0
        broken_lines_ratio = (
            1.0 - num_lines_endswith_period / (num_lines - num_empty_lines)
            if (num_lines - num_empty_lines) > 0
            else 0
        )
        if self.verbose:
            log.info(f"num_empty_lines={num_empty_lines}")
            log.info(f"num_lines_endswith_period={num_lines_endswith_period}")
            log.info(f"num_lines={num_lines}")
            log.info(f"empty_lines_ratio={empty_lines_ratio}")
            log.info(f"broken_lines_ratio={broken_lines_ratio}")
        if empty_lines_ratio > self._empty_lines_threshold:
            return True
        if broken_lines_ratio > self._broken_lines_threshold:
            return True

    def segment_article(self, article):

        if article is None:
            return None

        article = re.sub(
            f"{self._in_segment_separator}+", self._in_segment_separator, article
        )
        if self._merge_lines == "auto":
            merge_lines = self._check_if_merge_lines(article)
        else:
            merge_lines = self._merge_lines
        if merge_lines and self._merge_level == "article":
            if self.verbose > 10:
                log.info("=> merging lines of the article")
            article = re.sub(r"\s+", " ", article)

        if self._keep_segment:
            in_segments = article.split(self._in_segment_separator)
        else:
            in_segments = [article]

        segments = []
        for i, segment in enumerate(in_segments):
            segment = segment.strip()
            if len(segment) > 0:
                if self._filter_programming_language:
                    if self._skip_programming_language(segment):
                        continue
                if merge_lines and self._merge_level == "segment":
                    if self.verbose > 10:
                        log.info("=> merging lines of the segment")
                    segment = re.sub(r"\s+", " ", segment)

                if self._filter_language and self._detection_level == "segment":
                    if self._skip_language(segment):
                        continue
                sentences = []
                for sent in segment.split(self._in_sentence_separator):
                    sent = sent.strip()
                    if len(sent) > 0:

                        if (
                            self._filter_language
                            and self._detection_level == "sentence"
                        ):
                            if self._skip_language(sent):
                                continue

                        try:
                            sentences.extend(self._split_sentences(sent))
                        except Exception as e:
                            log.info(f"===> skipping segment {i}: {e}")
                            continue

                sentences = self.filter_sentence_length(sentences)
                if self._return_as_list:
                    segments.append(sentences)
                else:
                    segments.append(self._out_sentence_separator.join(sentences))

        return (
            segments
            if self._return_as_list
            else self._out_segment_separator.join(segments)
        )

    def filter_sentence_length(self, sentences):
        if self._filter_sentence_length:
            sentences = [
                sent
                for sent in sentences
                if len(sent.split()) >= self._min_num_words
                and len(sent) >= self._min_length
            ]
        return sentences

    def _skip_language(self, text):
        from ftlangdetect import detect

        lang = detect(text.replace("\n", " "), low_memory=False)
        if (
            lang is not None
            and lang["lang"] not in self._languages_to_keep
            and lang["score"] < self._min_language_probability
        ):
            if self.verbose > 5:
                log.info(f"===> lang={lang} not in languages_to_keep")
            return True
        return False

    # TODO: implement this
    # refer to: https://huggingface.co/huggingface/CodeBERTa-language-id
    def _skip_programming_language(self, text):
        return False
        # if self._guess is None:
        #     self._guess = Guess()

        # # Guess the language from code
        # lang = self._guess.language_name(text)
        # if lang is not None:
        #     if self.verbose > 5:
        #         log.info(f"===> Programming language: {lang}")
        #     if self.verbose > 10:
        #         log.info(text)
        #     return True
        # return False

    def _split_sentences(self, text):
        text = text.strip()
        if len(text) <= self._max_split_length:
            return self.segment(text)
        else:
            # if self.verbose:
            #     log.info(f"==> too long text: {len(text)}")
            text_chunk = text[: self._max_split_length]
            next_chunk = text[self._max_split_length :]
            sentences = []
            split_iter = 0
            while len(text_chunk) > 0 and split_iter < self._max_split_iterations:
                sents = self.segment(text_chunk)
                if len(sents) > 1 and len(next_chunk) > 0:
                    sentences.extend(sents[:-1])
                    last_sent = sents[-1]
                else:
                    sentences.extend(sents)
                    last_sent = ""
                end_pos = self._max_split_length - len(last_sent)
                if end_pos >= len(next_chunk):
                    text_chunk = last_sent + next_chunk
                    next_chunk = ""
                else:
                    text_chunk = last_sent + next_chunk[:end_pos]
                    next_chunk = next_chunk[end_pos:]
                split_iter += 1
            if split_iter >= self._max_split_iterations:
                if self.verbose:
                    log.critical(f"==> too many split iterations: {split_iter}")
                raise Exception(f"Maximum number of iterations reached: {split_iter}")
            return sentences

    def segment_articles(self, articles):
        return [self.segment_article(article) for article in articles]

    def chunk_article(self, article):

        if article is None:
            return None

        article = re.sub(
            f"{self._in_segment_separator}+", self._in_segment_separator, article
        )
        if self._keep_segment:
            in_segments = article.split(self._in_segment_separator)
        else:
            in_segments = [article]

        segments = []
        for i, segment in enumerate(in_segments):
            segment = segment.strip()
            segment = re.sub(
                f"{self._in_sentence_separator}+", self._in_sentence_separator, segment
            )
            if len(segment) > 0:
                sentences = segment.split(self._in_sentence_separator)

                chunks = self.chunk(
                    sentences,
                    max_length=self._chunk_size,
                    overlap=self._chunk_overlap,
                )
                if self._return_as_list != "list":
                    chunks = [
                        self._out_sentence_separator.join(chunk) for chunk in chunks
                    ]
                segments += chunks

        return (
            segments
            if self._return_as_list
            else self._out_segment_separator.join(segments)
        )

    def chunk(
        self,
        sentences: list,
        max_length: int,
        overlap: bool = False,
        len_func: Callable = None,
    ):
        """
        Split chunks from input texts by max_length.

        Args:
            text (Union[str, List[str], tuple]): input texts
            max_length (int): max length of ecah chunk
            overlap (bool): whether allow duplicated sentence

        Returns:
            chunks of segmented sentences
        """

        assert isinstance(sentences, list), "param `sentences` must be str."
        assert isinstance(max_length, int), "param `max_length` must be `int` type."
        assert isinstance(overlap, bool), "param `overlap` must be `bool` type."
        if len_func is None:
            len_func = self.len_func
        if not callable(len_func):
            len_func = len

        span, chunks = [], []

        for offset in _get_sentence_offsets(sentences, len_func):
            if len(span) > 0:
                if offset.this_cum_len - span[0].prev_cum_len > max_length:
                    chunks.append(_get_chunk_with_offset(sentences, span))
                    if overlap:
                        span = span[math.trunc(len(span) / 2) :]
                    else:
                        span = []

            span.append(offset)
        chunks.append(_get_chunk_with_offset(sentences, span))
        return chunks


def _get_sentence_offsets(sentences, len_func: Callable = None):
    sentence_offsets = []
    offset = None
    for i, sentence in enumerate(sentences):
        prev_cum_len = offset.this_cum_len if offset is not None else 0
        offset = Offset(i, i + 1, prev_cum_len, prev_cum_len + len_func(sentence))
        sentence_offsets.append(offset)
    return sentence_offsets


def _remove_useless_space(text):
    return re.sub(r"\s+", " ", text)


def _get_chunk_with_offset(sentences, span):
    start = span[0].this
    end = span[-1].next
    return sentences[start:end]


class NLTKSegmenter(Segmenter):
    def __init__(self, **kwargs):
        import nltk

        nltk.download("punkt", quiet=True)
        super().__init__(**kwargs)

    def segment(self, text):
        import nltk

        return nltk.tokenize.sent_tokenize(text)


class SimpleSegmenter(Segmenter):
    def __init__(self, simple={}, **kwargs):

        simple = simple or {}
        if not simple:
            simple["separator"] = "\n"
            simple["remove_newlines"] = True

        self.separator = codecs.decode(simple["separator"], "unicode_escape")
        self.remove_newlines = simple["remove_newlines"]
        log.info(f"segment_separator: -->{self.separator}<--")
        if self.remove_newlines:
            log.info("Remove newlines inside segments")
        super().__init__(**kwargs)

    def segment(self, text):
        segments = text.split(self.separator)
        if self.remove_newlines:
            segments = [segment.replace("\n", " ") for segment in segments]
        return segments


class PySBDSegmenter(Segmenter):
    def __init__(self, pysbd={}, **kwargs):
        import pysbd as pySBD

        pysbd = pysbd or {}
        if not pysbd:
            pysbd["language"] = "en"
            pysbd["clean"] = False

        if pysbd.get("doc_type", None) == "pdf":
            pysbd["clean"] = True
        self.seg = pySBD.Segmenter(**pysbd)
        super().__init__(**kwargs)

    def segment(self, text):
        return self.seg.segment(text)


class KSSSegmenter(Segmenter):
    def __init__(self, kss={}, **kwargs):
        from ekorpkit.preprocessors.kss import KSS

        kss = kss or {}
        if not kss:
            kss["use_heuristic"] = False
            kss["backend"] = "fugashi"

        self.seg = KSS(**kss)
        super().__init__(**kwargs)

    def segment(self, text):
        return self.seg.segment(text)



import logging
import os
import tempfile
from pathlib import Path

from .common import PathOrStr

logger = logging.getLogger(__name__)


class CacheFile:
    """
    This is a context manager that makes robust caching easier.

    On `__enter__`, an IO handle to a temporarily file is returned, which can
    be treated as if it's the actual cache file.

    On `__exit__`, the temporarily file is renamed to the cache file. If anything
    goes wrong while writing to the temporary file, it will be removed.
    """

    def __init__(self, cache_filename: PathOrStr, mode: str = "w+b", suffix: str = ".tmp") -> None:
        self.cache_filename = Path(cache_filename)
        self.cache_directory = os.path.dirname(self.cache_filename)
        self.mode = mode
        self.temp_file = tempfile.NamedTemporaryFile(
            self.mode, dir=self.cache_directory, delete=False, suffix=suffix
        )

    def __enter__(self):
        return self.temp_file

    def __exit__(self, exc_type, exc_value, traceback):
        self.temp_file.close()
        if exc_value is None:
            # Success.
            logger.debug(
                "Renaming temp file %s to cache at %s", self.temp_file.name, self.cache_filename
            )
            # Rename the temp file to the actual cache filename.
            os.replace(self.temp_file.name, self.cache_filename)
            return True
        # Something went wrong, remove the temp file.
        logger.debug("removing temp file %s", self.temp_file.name)
        os.remove(self.temp_file.name)
        return False



import os
import re
import logging
from statistics import mean
from scipy.stats import pearsonr, spearmanr
from transformers.data.metrics.squad_metrics import compute_exact, compute_f1
from .simple import SimpleTrainer


log = logging.getLogger(__name__)


class SimpleT5(SimpleTrainer):
    def __init__(self, **args):
        super().__init__(**args)

    def load_datasets(self):
        super().load_datasets()
        task_prefix = self.args.task_prefix._train_
        if task_prefix == self.Keys.CLASSIFICATION:
            train_data, _, _ = self.convert_to_train()
            self.labels_list = train_data[self.Keys.TARGET_TEXT].unique().tolist()
            log.info(f"Label list: {self.labels_list}")

    def convert_to_train(self):
        train_data, dev_data, test_data = super().convert_to_train()
        prefix_col = self._train_[self.Keys.PREFIX]
        task_prefix = self.args.task_prefix._train_
        if task_prefix is not None:
            train_data[prefix_col] = task_prefix
            if dev_data is not None:
                dev_data[prefix_col] = task_prefix
            if test_data is not None:
                test_data[prefix_col] = task_prefix
        return train_data, dev_data, test_data

    def train(self):
        from simpletransformers.t5 import T5Model

        args = self.args

        if not self.splits:
            self.load_datasets()
        train_data, dev_data, test_data = self.convert_to_train()
        if self.verbose:
            print(train_data.head())
        # Create a Model
        model = T5Model(
            args.model_type,
            args.model_uri,
            cuda_device=args.cuda_device,
            args=self._model_cfg,
        )

        # Train the model
        model.train_model(
            train_data,
            eval_data=dev_data,
        )

        # # Evaluate the model
        result = model.eval_model(test_data)
        print(result)

        # # Check predictions
        # # print(predictions[:5])
        # return result, model_outputs, predictions

    def load_model(self, model_dir=None, pred_args=None):
        from simpletransformers.t5 import T5Model

        if model_dir is None:
            model_dir = self.args.config.best_model_dir
        if pred_args is not None:
            self._model_cfg.update(pred_args)

        self.model = T5Model(self.args.model_type, model_dir, args=self._model_cfg)
        log.info(f"Loaded model from {model_dir}")

    def _predict(self, data: list):
        if self.model is None:
            self.load_model()

        preds = self.model.predict(data)
        return preds

    def convert_to_predict(self, data):
        input_col = self._predict_[self.Keys.INPUT]
        prefix_col = self._predict_[self.Keys.PREFIX]
        task_prefix = self.args.task_prefix._predict_
        if task_prefix is not None:
            data[prefix_col] = task_prefix

        data_to_predict = [
            prefix + ": " + str(input_text)
            for prefix, input_text in zip(
                data[prefix_col].tolist(), data[input_col].tolist()
            )
        ]

        if self.verbose:
            print(data_to_predict[:5])
        return data_to_predict

    def append_predictions(self, df, preds):
        predicted_col = self._predict_[self.Keys.PREDICTED]
        if self.args.config.num_return_sequences > 1:
            preds = [pred[0] for pred in preds]
        df[predicted_col] = preds
        return df


def f1(truths, preds):
    return mean([compute_f1(truth, pred) for truth, pred in zip(truths, preds)])


def exact(truths, preds):
    return mean([compute_exact(truth, pred) for truth, pred in zip(truths, preds)])


def pearson_corr(preds, labels):
    return pearsonr(preds, labels)[0]


def spearman_corr(preds, labels):
    return spearmanr(preds, labels)[0]


def convert_BIO_labels(filename):
    result_labels = []
    with open(filename, "r", encoding="utf-8") as file:
        cnt = 0
        for line in file:
            line = re.sub(r"\*(\w+)", r"\1*", line)
            tokens = re.sub(
                r'[!"#$%&\'()+,-.:;<=>?@[\\\]^_`{\|}~⁇]', " ", line.strip()
            ).split()
            seq_label = []
            start_entity = 0
            entity_type = "O"
            for idx, token in enumerate(tokens):
                if token.endswith("*"):
                    start_entity += (
                        1 if (start_entity == 0 or token[:-1] != entity_type) else -1
                    )
                    entity_type = token[:-1]
                else:
                    if start_entity == 0:
                        seq_label.append("O")
                        entity_type = "O"
                    elif start_entity < 0:
                        raise "Something errors"
                    else:
                        if tokens[idx - 1].endswith("*"):
                            seq_label.append("B-" + entity_type.upper())
                        else:
                            seq_label.append("I-" + entity_type.upper())

            result_labels.append(seq_label)
            cnt += 1
    #             if cnt % 100 == 0:
    #                 print('Processed %d sentences' % cnt)
    return result_labels



import codecs
import simdjson
import orjson as json
import jsonpath_ng as jpath
import logging
from hydra.utils import instantiate
from omegaconf.listconfig import ListConfig
from ekorpkit.utils.func import any_to_utf8
from tqdm.auto import tqdm
from joblib import Parallel, delayed
from ekorpkit.utils.batch.batcher import tqdm_joblib


log = logging.getLogger(__name__)


parser = simdjson.Parser()


def _load_with_parser(parser, contents):
    if contents is None:
        return None
    contents = instantiate(parser, contents)
    return contents
    # try:
    # 	contents = instantiate(parser, contents)
    # 	return contents
    # except Exception:
    # 	msg.fail('Error parsing with parser {}'.format(parser))
    # 	# print(f'Error parsing {contents}')
    # 	return None


def json_loads(contents):
    try:
        return parser.parse(contents).as_dict()
    except ValueError:
        return contents


def load_json_to_list(contents, jpath_expression):
    contents = json.loads(contents)
    jpath_expression = jpath.parse(jpath_expression)
    results = []
    for match in jpath_expression.find(contents):
        results.append(match.value)
    return results


def load_json(contents):
    if contents is None:
        return None
    if len(contents) == 0:
        return None
    try:
        return [json.loads(contents)]
    except json.JSONDecodeError:
        log.critical(f"Error parsing {contents}")
        raise Exception


def load_jsonlines(contents):
    lines = contents.split("\n")
    # print(f'Loading jsonlines from {len(lines)} lines')
    data = []
    for lineno, line in enumerate(lines):
        if len(line) == 0:
            continue
        js = None
        try:
            js = json.loads(line)
        except json.JSONDecodeError as err:
            log.warning(f"Skipping lineno {lineno} - colno:{err.colno} msg:{err.msg}")
        if js:
            data.append(js)

    return data


def parse_data(contents, parse_args, default_items, filename="", num_workers=1):
    parser = parse_args.get("parser", None)
    decompressor = parse_args.get("decompressor", None)
    default_items = default_items.copy()
    if filename is not None and len(filename) > 0:
        default_items["filename"] = filename

    if decompressor is not None and ".gz" in filename:
        contents = instantiate(decompressor, contents)

    decode_before_parse = parse_args.get("decode_before_parse", False)
    multiprocessing_at = parse_args.get("multiprocessing_at", "load_data")

    documents = []
    if decode_before_parse and isinstance(contents, (bytes, bytearray)):
        try:
            contents = any_to_utf8(contents)
            # contents = ftfy.fix_text(any_to_utf8(contents))
        except Exception:
            log.warning(f"Decoding error")
            print(contents)
            return documents

    if parser is not None and parser.get("_target_", None):
        contents = _load_with_parser(parser, contents)
    else:
        contents = load_json(contents)
    if contents is None:
        return documents
    if len(contents) == 0:
        return documents

    if isinstance(contents, str):
        contents = [contents]

    if num_workers > 1 and multiprocessing_at == "parse_data":
        log.info(
            f"Starting multiprocessing with {num_workers} processes at {multiprocessing_at}"
        )
        desciption = default_items.get("filename", "::parse_data()")
        with tqdm_joblib(tqdm(desc=desciption, total=len(contents))) as pbar:
            results = Parallel(n_jobs=num_workers)(
                delayed(parse_json)(content, parse_args, default_items, 1)
                for content in contents
                if len(content) > 0
            )
            for result in results:
                documents += result
    else:
        if len(contents) > 1000:
            log.info("Number of data in the contents: {}".format(len(contents)))
        for i, content in enumerate(contents):
            # print(content)
            if len(content) > 0:
                documents += parse_json(content, parse_args, default_items, num_workers)

    return documents


def parse_json(data, parse_args, default_items={}, num_workers=1):

    documents = []
    meta_args = parse_args.get("meta", {})
    meta_field = meta_args.get("field", None)
    data_args = parse_args["data"]
    data_field = data_args.get("field", None)
    default_items = default_items.copy()

    if meta_args.get("item", None) is not None:
        meta_data = data[meta_field] if meta_field else data
        meta = data_to_document(meta_data, meta_args)
        if len(meta) > 0:
            default_items.update(meta[0])

    data = data[data_field] if data_field else data
    if isinstance(data, list):
        totalcount = len(data)
        if num_workers > 1 and totalcount > num_workers:
            log.info(
                f"Starting multiprocessing with {num_workers} processes at parse_json"
            )
            desciption = default_items.get("filename", "::parse_json()")
            with tqdm_joblib(tqdm(desc=desciption, total=totalcount)) as pbar:
                results = Parallel(n_jobs=num_workers)(
                    delayed(data_to_document)(doc, data_args, default_items)
                    for doc in data
                )
                for result in results:
                    documents += result
        else:
            if totalcount > 1000:
                log.info("Total number of documents: {}".format(totalcount))
            for ix, doc in enumerate(data):
                documents += data_to_document(doc, data_args, default_items)
    else:
        documents += data_to_document(data, data_args, default_items)

    return documents


def data_to_document(data, data_args, default_items={}):
    documents = []
    # data.update(default_items)

    item_separator = data_args.get("item_separator", "")
    item_separator = codecs.decode(item_separator, "unicode_escape")

    document = {}
    for k, v in data_args["item"].items():
        val = get_item_value(data, v)
        if isinstance(val, list):
            strval = []
            for sv in val:
                if sv:
                    if len(str(sv).strip()) > 0:
                        strval.append(str(sv).strip())
                # else:
                # 	# strval.append('')
                # 	print(v)
            val = item_separator.join(strval)
        document[k] = val
    document.update(default_items)
    documents.append(document)
    return documents


def get_item_value(item, key):
    if item is None or key is None:
        return None
    try:
        if isinstance(key, str):
            if "." in key:
                return get_value_by_jpath(item, key)
            else:
                return item[key] if key in item else None
        elif isinstance(key, (list, ListConfig)):
            value = []
            for k in key:
                val = get_item_value(item, k)
                if isinstance(val, list):
                    value += val
                else:
                    value.append(val)
            return value
        else:
            print(key, type(key))
            raise ValueError("key must be str or list")
    except Exception as e:
        log.critical(item)
        log.critical(key, type(key))
        print(e)
        raise e
        # return None


def get_value_by_jpath(item, expression):
    jpath_expression = jpath.parse(expression)
    results = []
    for match in jpath_expression.find(item):
        if isinstance(match.value, list):
            val = "\n".join(match.value)
        else:
            val = match.value
        results.append(val)
    if len(results) == 0:
        return None
    elif len(results) == 1:
        return results[0]
    else:
        return results



import gzip
import shutil
import os
import tarfile
import zipfile
from tqdm.auto import tqdm
from urllib import request
from ekorpkit import eKonf


def download_from_gcs(**cfg):
    """Download the data from source url, unless it's already here.

    Args:
        filename: string, name of the file in the directory.
        work_directory: string, path to working directory.
        source_url: url to download from if file doesn't exist.

    Returns:
        Path to resulting file.
    """
    import tensorflow as tf

    args = eKonf.to_config(cfg)
    os.makedirs(args.output_dir, exist_ok=True)
    if not os.listdir(args.output_dir) or args.force_download:
        for filename, source_path in args.data_sources.items():
            output_path = f"{args.output_dir}/{filename}"
            print(f"Downloading {source_path} to {output_path}")
            tf.io.gfile.copy(source_path, output_path, overwrite=args.force_download)
    else:
        print(f" >> {args.output_dir} is not empty, skip downloading")


def download_from_web(**cfg):
    args = eKonf.to_config(cfg)
    os.makedirs(args.output_dir, exist_ok=True)
    if not os.listdir(args.output_dir) or args.force_download:
        for filename, url in args.urls.items():
            filepath = f"{args.output_dir}/{filename}"
            print(f"Downloading {url} to {filepath}")
            # download_file(url, file_path)
            web_download(url, filepath, args.name, force_download=False)
        if args.extract_command:
            print(args.extract_command)
    else:
        print(f" >> {args.output_dir} is not empty, skip downloading")


def download_from_gdrive(**cfg):
    args = eKonf.to_config(cfg)
    os.makedirs(args.output_dir, exist_ok=True)
    for filename, url in args.urls.items():
        file_path = f"{args.output_dir}/{filename}"
        gdrive_download(url, file_path, args.name, args.force_download)


def download_from_gdrive_untar(**cfg):
    args = eKonf.to_config(cfg)
    os.makedirs(args.output_dir, exist_ok=True)
    for filename, url in args.urls.items():
        file_path = f"{args.output_dir}/{filename}"
        gdrive_download_untar(url, file_path, args.name, args.force_download)


def web_download(
    url, local_path, filename="", force_download=False, verbose=True, **kwargs
):
    if filename == "":
        filename = os.path.basename(local_path)
    site = request.urlopen(url)
    meta = site.info()
    remote_size = int(meta["Content-Length"])
    if (
        (not force_download)
        and os.path.exists(local_path)
        and (os.stat(local_path).st_size == remote_size)
    ):
        if verbose:
            print(f"[{filename}] is already downloaded at {local_path}")
        return None
    filename = os.path.basename(local_path)
    if not os.path.exists(os.path.dirname(local_path)):
        os.makedirs(os.path.dirname(local_path), exist_ok=True)
    with tqdm(
        unit="B", unit_scale=True, miniters=1, desc=f"[{filename}] download {filename}"
    ) as t:
        request.urlretrieve(url, filename=local_path, reporthook=_reporthook(t))


def web_download_unzip(
    url, zip_path, filename="", force_download=False, verbose=True, **kwargs
):
    web_download(url, zip_path, filename, force_download)
    # assume that path/to/abc.zip consists path/to/abc
    data_path = zip_path[:-4]
    if (not force_download) and os.path.exists(data_path):
        if verbose:
            print(f"[{filename}] is already extracted at {data_path}")
        return None
    data_root = os.path.dirname(zip_path)
    with zipfile.ZipFile(zip_path, "r") as zip_ref:
        zip_ref.extractall(data_root)
    if verbose:
        print(f"unzip {data_path}")


def web_download_un7z(
    url, zip_path, filename="", force_download=False, verbose=True, **kwargs
):
    import py7zr

    web_download(url, zip_path, filename, force_download)
    # assume that path/to/abc.zip consists path/to/abc
    data_path = zip_path[:-3]
    if (not force_download) and os.path.exists(data_path):
        if verbose:
            print(f"[{filename}] is already extracted at {data_path}")
        return None
    # data_root = os.path.dirname(zip_path)
    with py7zr.SevenZipFile(zip_path, mode="r") as z:
        z.extractall()
    if verbose:
        print(f"un7z {data_path}")


def web_download_untar(
    url, tar_path, filename="", force_download=False, verbose=True, **kwargs
):
    web_download(url, tar_path, filename, force_download)
    # assume that path/to/abc.tar consists path/to/abc
    data_path = tar_path[:-4]
    if (not force_download) and os.path.exists(data_path):
        if verbose:
            print(f"[{filename}] is already extracted at {data_path}")
        return None
    data_root = os.path.dirname(tar_path)
    with tarfile.open(tar_path) as tar:
        tar.extractall(data_root)
    if verbose:
        print(f"decompress {tar_path}")


def web_download_ungzip(
    url, gzip_path, filename="", force_download=False, verbose=True, **kwargs
):
    web_download(url, gzip_path, filename, force_download)
    # assume that path/to/abc.gzip consists path/to/abc
    data_path = gzip_path[:-3]
    if (not force_download) and os.path.exists(data_path):
        if verbose:
            print(f"[{filename}] is already extracted at {data_path}")
        return None
    with gzip.open(gzip_path, "rb") as fi:
        with open(data_path, "wb") as fo:
            shutil.copyfileobj(fi, fo)
    print(f"decompress {gzip_path}")


def gdrive_download_un7z(
    file_id, zip_path, filename="", force_download=False, verbose=True, **kwargs
):
    import py7zr

    gdrive_download(file_id, zip_path, filename, force_download)
    # assume that path/to/abc.zip consists path/to/abc
    data_path = zip_path[:-3]
    if (not force_download) and os.path.exists(data_path):
        if verbose:
            print(f"[{filename}] is already extracted at {data_path}")
        return None
    # data_root = os.path.dirname(zip_path)
    with py7zr.SevenZipFile(zip_path, mode="r") as z:
        z.extractall(path=data_path)
    if verbose:
        print(f"un7z {data_path}")


def gdrive_download_untar(
    file_id, local_path, filename="", force_download=False, verbose=True, **kwargs
):
    gdrive_download(file_id, local_path, filename, force_download)
    # assume that path/to/abc.tar consists path/to/abc
    data_path = local_path[: local_path.find(".tar")]
    if (not force_download) and os.path.exists(data_path):
        if verbose:
            print(f"[{filename}] is already extracted at {data_path}")
        return None
    data_root = os.path.dirname(local_path)
    with tarfile.open(local_path) as tar:
        tar.extractall(data_root)
    if verbose:
        print(f"decompress {local_path}")


def gdrive_download(
    file_id,
    local_path,
    filename="",
    force_download=False,
    verify=True,
    verbose=True,
    **kwargs,
):
    import gdown

    if filename == "":
        filename = os.path.basename(local_path)
    if (not force_download) and os.path.exists(local_path):
        if verbose:
            print(f"[ekorpkit] [{filename}] is already downloaded at {local_path}")
        return None

    if os.path.dirname(local_path):
        os.makedirs(os.path.dirname(local_path), exist_ok=True)

    gdown.download(
        id=file_id,
        output=local_path,
        quiet=not verbose,
        fuzzy=True,
        verify=verify,
    )
    if verbose:
        print(f"download {filename}")


def _reporthook(t):
    """``reporthook`` to use with ``urllib.request`` that prints the process of the download.

    Uses ``tqdm`` for progress bar.

    **Reference:**
    https://github.com/tqdm/tqdm

    Args:
        t (tqdm.tqdm) Progress bar.

    Example:
        >>> with tqdm(unit='B', unit_scale=True, miniters=1, desc=filename) as t:  # doctest: +SKIP
        ...   urllib.request.urlretrieve(file_url, filename=full_path, reporthook=reporthook(t))
    """
    last_b = [0]

    def inner(b=1, bsize=1, tsize=None):
        """
        Args:
            b (int, optional): Number of blocks just transferred [default: 1].
            bsize (int, optional): Size of each block (in tqdm units) [default: 1].
            tsize (int, optional): Total size (in tqdm units). If [default: None] remains unchanged.
        """
        if tsize is not None:
            t.total = tsize
        t.update((b - last_b[0]) * bsize)
        last_b[0] = b

    return inner



"""
Google Cloud Storage.
"""

import io
from typing import Optional, Tuple
from google.cloud import storage
from ..common import _split_cloud_path
from .scheme_client import SchemeClient


class GsClient(SchemeClient):
    scheme = "gs"

    def __init__(self, resource: str) -> None:
        super().__init__(resource)
        self.blob = GsClient.get_gcs_blob(resource)
        self._loaded = False

    def load(self):
        from google.api_core.exceptions import NotFound

        if not self._loaded:
            try:
                self.blob.reload()
                self._loaded = True
            except NotFound:
                raise FileNotFoundError(self.resource)

    def get_etag(self) -> Optional[str]:
        self.load()
        return self.blob.etag or self.blob.md5_hash

    def get_size(self) -> Optional[int]:
        self.load()
        return self.blob.size

    def get_resource(self, temp_file: io.BufferedWriter) -> None:
        from google.cloud.storage.retry import DEFAULT_RETRY

        self.load()
        self.blob.download_to_file(temp_file, checksum="md5", retry=DEFAULT_RETRY)

    @staticmethod
    def split_gcs_path(resource: str) -> Tuple[str, str]:
        return _split_cloud_path(resource, "gs")

    @staticmethod
    def get_gcs_blob(resource: str) -> storage.blob.Blob:
        from google.auth.exceptions import DefaultCredentialsError

        try:
            gcs_resource = storage.Client()
        except DefaultCredentialsError:
            gcs_resource = storage.Client.create_anonymous_client()
        bucket_name, gcs_path = GsClient.split_gcs_path(resource)
        bucket = gcs_resource.bucket(bucket_name)
        blob = bucket.blob(gcs_path)
        return blob



from fileinput import filename
import logging
import logging
import os
import pandas as pd
from ekorpkit import eKonf


log = logging.getLogger(__name__)


class Quandl:
    def __init__(self, **args):
        from fredapi import Fred
        import nasdaqdatalink

        self.args = eKonf.to_dict(args)
        self.autoload = self.args.get("autoload", True)
        self.name = self.args.get("name")
        self.fred_api_key = self.args.get("fred_api_key")
        self.nasdaq_api_key = self.args.get("nasdaq_api_key")
        self.verbose = self.args.get("verbose", True)
        self.series_id = self.args.get("series_id")
        if isinstance(self.series_id, str):
            self.series_id = [self.series_id]
        elif not isinstance(self.series_id, list):
            self.series_id = []
        self.series_name = self.args.get("series_name")
        self.value_column = self.args.get("value_column") or "value"
        if self.series_name is None:
            if self.series_id:
                self.series_name = "_".join(self.series_id).replace("/", "_")
            else:
                self.series_name = self.value_column

        self.start_date = self.args.get("start_date")
        self.end_date = self.args.get("end_date")
        self.eval_columns = self.args.get("pipeline").get("eval_columns")

        self.output_dir = self.args["output_dir"]
        os.makedirs(self.output_dir, exist_ok=True)
        self.output_file = self.args["output_file"]
        self.force_download = self.args["force_download"]

        self.fred = Fred(api_key=self.fred_api_key)
        nasdaqdatalink.ApiConfig.api_key = self.nasdaq_api_key
        self.data = None

        if self.autoload:
            self.load(
                series_id=self.series_id,
                series_name=self.series_name,
                start_date=self.start_date,
                end_date=self.end_date,
            )

    def get(self, series_id, start_date=None, end_date=None, **kwargs):
        if "/" in series_id:
            return self.get_nasqaq(series_id, start_date, end_date, **kwargs)
        else:
            series = self.get_series(series_id, start_date, end_date, **kwargs)
            df = pd.DataFrame(series, columns=[self.value_column])
            return df

    def get_nasqaq(self, series_id, start_date=None, end_date=None, **kwargs):
        """Return dataframe of requested dataset from Nasdaq Data Link.
        :param series_id: str or list, depending on single dataset usage or multiset usage
                Dataset codes are available on the Nasdaq Data Link website
        :param str api_key: Downloads are limited to 50 unless api_key is specified
        :param str start_date, end_date: Optional datefilers, otherwise entire
            dataset is returned
        :param str collapse: Options are daily, weekly, monthly, quarterly, annual
        :param str transform: options are diff, rdiff, cumul, and normalize
        :param int rows: Number of rows which will be returned
        :param str order: options are asc, desc. Default: `asc`
        :param str returns: specify what format you wish your dataset returned as,
            either `numpy` for a numpy ndarray or `pandas`. Default: `pandas`
        :returns: :class:`pandas.DataFrame` or :class:`numpy.ndarray`
        Note that Pandas expects timeseries data to be sorted ascending for most
        timeseries functionality to work.
        Any other `kwargs` passed to `get` are sent as field/value params to Nasdaq Data Link
        with no interference.
        """
        import nasdaqdatalink

        return nasdaqdatalink.get(
            dataset=series_id, start_date=start_date, end_date=end_date, **kwargs
        )

    def get_series(self, series_id, start_date=None, end_date=None, **kwargs):
        """
        Get data for a Fred series id. This fetches the latest known data, and is equivalent to get_series_latest_release()

        Parameters
        ----------
        series_id : str
            Fred series id such as 'CPIAUCSL'
        start_date : datetime or datetime-like str such as '7/1/2014', optional
            earliest observation date
        end_date : datetime or datetime-like str such as '7/1/2014', optional
            latest observation date
        kwargs : additional parameters
            Any additional parameters supported by FRED. You can see https://api.stlouisfed.org/docs/fred/series_observations.html for the full list

        Returns
        -------
        data : Series
            a Series where each index is the observation date and the value is the data for the Fred series
        """
        return self.fred.get_series(
            series_id,
            observation_start=start_date,
            observation_end=end_date,
            **kwargs,
        )

    def load(
        self,
        series_id,
        series_name=None,
        start_date=None,
        end_date=None,
        filename=None,
        expressions=None,
        index_name="date",
        reset_index=False,
    ):
        if isinstance(series_id, str):
            series_id = [series_id]
        elif not isinstance(series_id, list):
            series_id = []

        if series_name is None:
            series_name = "_".join(series_id).replace("/", "_")
        if filename is None:
            filename = f"{series_name}.parquet"

        if start_date is None:
            start_date = self.start_date
        if end_date is None:
            end_date = self.end_date

        if isinstance(self.eval_columns, dict):
            self.eval_columns["expressions"] = expressions

        if self.verbose:
            print(f"Loading {series_name}{series_id} from {start_date} to {end_date}")

        filepath = os.path.join(self.output_dir, filename)
        if not os.path.exists(filepath) or self.force_download:
            self.data = self._load_series(
                series_id, series_name, start_date, end_date, index_name, reset_index
            )
            eKonf.save_data(self.data, filepath, verbose=self.verbose)
        else:
            log.info(f"{filepath} already exists.")
            self.data = eKonf.load_data(filepath, verbose=self.verbose)
        return self.data.copy()

    def _load_series(
        self,
        series_ids=None,
        series_name=None,
        start_date=None,
        end_date=None,
        index_name="date",
        reset_index=False,
        **kwargs,
    ):

        _dfs = []
        for series_id in series_ids:
            df = self.get(
                series_id=series_id,
                start_date=start_date,
                end_date=end_date,
                **kwargs,
            )
            if len(df.columns) == 1:
                if series_name is None:
                    series_name = df.columns[0]
                df.columns = [self.value_column]
            df = eKonf.pipe(df, self.eval_columns)
            if len(series_ids) > 1:
                df["series_id"] = series_id
            if series_name:
                columns = {
                    col: col.replace(self.value_column, series_name)
                    for col in df.columns
                    if col.startswith(self.value_column)
                }
                df.rename(columns=columns, inplace=True)
            if self.verbose:
                print(df.head())
            _dfs.append(df)

        df = pd.concat(_dfs)
        df.index.name = index_name
        if reset_index:
            df.index.name = series_name + "_" if series_name else "" + df.index.name
            df = df.reset_index()
        return df



from collections import Counter
from typing import List


def check_unique_names(names: List[str]) -> None:
    """Check that operator names are unique."""
    k, ct = Counter(names).most_common(1)[0]
    if ct > 1:
        raise ValueError(f"Operator names not unique: {ct} operators with name {k}")



import logging
import re
import threading
import os
import numpy as np
import pandas as pd
import codecs
import requests
from tqdm.auto import tqdm
from bs4 import BeautifulSoup
from datetime import datetime
from abc import abstractmethod
from ekorpkit import eKonf


log = logging.getLogger(__name__)


def download_data(fomc, from_year, verbose=False):
    if not os.path.exists(fomc.output_filepath) or fomc.force_download:
        log.info(f"Downloading {fomc.content_type}")
        df = fomc.get_contents(from_year)
        log.info("Shape of the downloaded data: ", df.shape)
        if verbose:
            print("The first 5 rows of the data: \n", df.head())
            print("The last 5 rows of the data: \n", df.tail())
        fomc.save()
    else:
        log.info(f"{fomc.content_type} already exists")


def build_fomc(**args):

    args = eKonf.to_config(args)
    from_year = args.from_year
    verbose = args.verbose
    if (from_year < 1980) or (from_year > 2020):
        log.warning("Please specify the second argument between 1980 and 2020")
        return

    for content in args.contents:
        fomc = eKonf.instantiate(content, **args.fomc)
        download_data(fomc, from_year, verbose)


class FOMC:
    """
    A base class for extracting documents from the FOMC website
    """

    def __init__(self, content_type, **args):
        args = eKonf.to_dict(args)
        self.verbose = args["verbose"]
        log.info(f"Initializing {content_type}")

        # Set arguments to internal variables
        self.content_type = content_type
        self.num_workers = args["num_workers"]
        self.output_dir = args["output_dir"]
        self.output_raw_dir = os.path.join(self.output_dir, "raw")
        os.makedirs(self.output_raw_dir, exist_ok=True)
        self.output_file = content_type + ".parquet"
        self.output_filepath = os.path.join(self.output_dir, self.output_file)
        self.calendar_filepath = os.path.join(self.output_dir, "fomc_calendar.parquet")

        self.segment_separator = codecs.decode(
            args["segment_separator"], "unicode_escape"
        )
        self.force_download = args["force_download"]

        # Initialization
        self.df = None
        self.links = None
        self.dates = None
        self.articles = None
        self.speakers = None
        self.titles = None

        # missing articles
        self._articles = args.get("articles")

        # FOMC website URLs
        self.base_url = args["base_url"]
        self.calendar_url = args["calendar_url"]

        _meta = args["meta"]
        # FOMC Chairperson's list
        self.chairpersons = _meta.get("chairpersons")
        if isinstance(self.chairpersons, dict):
            self.chairpersons = pd.DataFrame(self.chairpersons.values())
        self.canlendar = None

        self.econ_series = _meta.get("econ_series")
        self.recessions = _meta.get("recessions")
        if isinstance(self.recessions, dict):
            self.recessions = pd.DataFrame(self.recessions.values())

        self.unconventionals = _meta.get("unconventionals")
        if isinstance(self.unconventionals, dict):
            self.unconventionals = pd.DataFrame(self.unconventionals.values())
            self.unconventionals["speaker"] = self.unconventionals.date.map(
                self._speaker_from_date
            )

        # skip list
        self.skip_text_list = [
            "Return to top",
        ]

    def _date_from_link(self, link):
        # TODO: fix this
        date = re.findall("[0-9]{8}", link)[0]
        if date[4] == "0":
            date = "{}-{}-{}".format(date[:4], date[5:6], date[6:])
        else:
            date = "{}-{}-{}".format(date[:4], date[4:6], date[6:])
        return date

    def _speaker_from_date(self, article_date):
        """
        Returns the speaker of the article based on the date of the article
        """
        if isinstance(article_date, str):
            article_date = datetime.strptime(article_date, "%Y-%m-%d")

        speaker = self.chairpersons.query(
            "from_date < @article_date & to_date > @article_date"
        )
        if speaker.empty:
            return "other"
        else:
            speaker = speaker.iloc[0]
            return speaker.first_name + " " + speaker.last_name

    @abstractmethod
    def _get_links(self, from_year):
        """
        private function that sets all the links for the FOMC meetings
        from the giving from_year to the current most recent year
        from_year is min(2015, from_year)
        """
        # Implement in sub classes
        # TODO - replace links list with dictionary with key as date
        pass

    @abstractmethod
    def _add_article(self, link, index=None):
        """
        adds the related article for 1 link into the instance variable
        index is the index in the article to add to. Due to concurrent
        prcessing, we need to make sure the articles are stored in the
        right order
        """
        # Implement in sub classes
        pass

    def _get_articles_multi(self):
        """
        gets all articles using multi-threading
        """
        if self.verbose:
            print("Getting articles - Multi-threaded...")

        self.articles = [""] * len(self.links)
        jobs = []
        # initiate and start threads:
        index = 0
        while index < len(self.links):
            if len(jobs) < self.num_workers:
                t = threading.Thread(
                    target=self._add_article,
                    args=(
                        self.links[index],
                        index,
                    ),
                )
                jobs.append(t)
                t.start()
                index += 1
            else:  # wait for threads to complete and join them back into the main thread
                t = jobs.pop(0)
                t.join()
        for t in jobs:
            t.join()

        # for row in range(len(self.articles)):
        #    self.articles[row] = self.articles[row].strip()

    def get_contents(self, from_year=1990):
        """
        Returns a Pandas DataFrame with the date as the index for a date range of from_year to the most current.
        Save the same to internal df as well.
        """
        self._get_links(from_year)
        self._get_articles_multi()
        dict = {
            "date": self.dates,
            "speaker": self.speakers,
            "title": self.titles,
            "text": self.articles,
        }
        self.df = pd.DataFrame(dict)
        self.df["content_type"] = self.content_type
        _df = self.get_missing_contents(self.content_type)
        if _df is not None:
            self.df = pd.concat([self.df, _df])
        self.df.sort_values(by="date", inplace=True)
        self.df.reset_index(drop=True, inplace=True)
        return self.df

    def get_missing_contents(self, content_type):
        if isinstance(self._articles, dict) and content_type in self._articles:
            articles = self._articles[content_type]
            df = pd.DataFrame.from_dict(articles, orient="records")
            df["content_type"] = content_type
            return df
        else:
            return None

    def save(self):
        """
        save the dataframe to a parquet file
        """
        log.info("Writing to ", self.output_filepath)
        eKonf.save_data(self.df, self.output_filepath)

    def load_calendar(self, from_year=None, force_download=False):
        """
        get the calendar from the FOMC website
        """
        if os.path.exists(self.calendar_filepath) and not force_download:
            log.info("Loading calendar from cache...")
            self.calendar = eKonf.load_data(self.calendar_filepath)
            return self.calendar

        if self.verbose:
            print("Getting calendar...")

        if from_year:
            from_year = int(from_year)

            if (from_year < 1936) or (from_year > 2017):
                print("Please specify the first argument between 1936 and 2015")
                return
        else:
            from_year = 1936
            print(
                "From year is set as 1936. Please specify the year as the first argument if required."
            )

        fomc_dates = []
        # Retrieve FOMC Meeting date from current page - recent five years
        r = requests.get(self.calendar_url)
        soup = BeautifulSoup(r.text, "html.parser")
        panel_divs = soup.find_all("div", {"class": "panel panel-default"})

        for panel_div in panel_divs:
            m_year = panel_div.find("h4").get_text()[:4]
            m_months = panel_div.find_all("div", {"class": "fomc-meeting__month"})
            m_dates = panel_div.find_all("div", {"class": "fomc-meeting__date"})
            if self.verbose:
                print("YEAR: {} - {} meetings found.".format(m_year, len(m_dates)))

            for (m_month, m_date) in zip(m_months, m_dates):
                month_name = m_month.get_text().strip()
                date_text = m_date.get_text().strip()
                is_forecast = False
                is_unscheduled = False
                is_month_short = False

                if "cancelled" in date_text:
                    continue
                elif "notation vote" in date_text:
                    date_text = date_text.replace("(notation vote)", "").strip()
                elif "unscheduled" in date_text:
                    date_text = date_text.replace("(unscheduled)", "").strip()
                    is_unscheduled = True

                if "*" in date_text:
                    date_text = date_text.replace("*", "").strip()
                    is_forecast = True

                if "/" in month_name:
                    month_name = re.findall(r".+/(.+)$", month_name)[0]
                    is_month_short = True

                if "-" in date_text:
                    date_text = re.findall(r".+-(.+)$", date_text)[0]

                meeting_date_str = m_year + "-" + month_name + "-" + date_text
                if is_month_short:
                    meeting_date = datetime.strptime(meeting_date_str, "%Y-%b-%d")
                else:
                    meeting_date = datetime.strptime(meeting_date_str, "%Y-%B-%d")

                fomc_dates.append(
                    {
                        "date": meeting_date,
                        "unscheduled": is_unscheduled,
                        "forecast": is_forecast,
                        "confcall": False,
                    }
                )
        min_recent_year = min(fomc_dates, key=lambda x: x["date"]).get("date").year
        # Retrieve FOMC Meeting date older than 2015
        for year in range(from_year, min_recent_year):
            hist_url = self.base_url + f"/monetarypolicy/fomchistorical{year}.htm"
            r = requests.get(hist_url)
            soup = BeautifulSoup(r.text, "html.parser")
            if year >= 2011:
                panel_headings = soup.find_all("h5", {"class": "panel-heading"})
            else:
                panel_headings = soup.find_all("div", {"class": "panel-heading"})
            log.info("YEAR: {} - {} meetings found.".format(year, len(panel_headings)))
            for panel_heading in panel_headings:
                date_text = panel_heading.get_text().strip()
                # print("Date: ", date_text)
                regex = r"(January|February|March|April|May|June|July|August|September|October|November|December).*\s(\d*-)*(\d+)\s+(Meeting|Conference Calls?|\(unscheduled\))\s-\s(\d+)"
                date_text_ext = re.findall(regex, date_text)[0]
                meeting_date_str = (
                    date_text_ext[4] + "-" + date_text_ext[0] + "-" + date_text_ext[2]
                )
                # print("   Extracted:", meeting_date_str)
                if meeting_date_str == "1992-June-1":
                    meeting_date_str = "1992-July-1"
                elif meeting_date_str == "1995-January-1":
                    meeting_date_str = "1995-February-1"
                elif meeting_date_str == "1998-June-1":
                    meeting_date_str = "1998-July-1"
                elif meeting_date_str == "2012-July-1":
                    meeting_date_str = "2012-August-1"
                elif meeting_date_str == "2013-April-1":
                    meeting_date_str = "2013-May-1"

                meeting_date = datetime.strptime(meeting_date_str, "%Y-%B-%d")
                is_confcall = "Conference Call" in date_text_ext[3]
                is_unscheduled = "unscheduled" in date_text_ext[3]
                fomc_dates.append(
                    {
                        "date": meeting_date,
                        "unscheduled": is_unscheduled,
                        "forecast": False,
                        "confcall": is_confcall,
                    }
                )

        df = pd.DataFrame(fomc_dates).sort_values(by=["date"])

        df["speaker"] = df.date.map(self._speaker_from_date)

        # # Use date as index
        df.set_index("date", inplace=True)

        eKonf.save_data(df, self.calendar_filepath)
        self.calendar = df
        return df.copy()

    def add_decisions_to_calendar(self, fedrates, series_id="DFEDTAR"):
        """
        The target range was changed a couple of days after the announcement in the past,
         while it is immediately put in effect on the day recently.
        Use the target rate three days after the meeting as target announced,
         compare it with previous day's rate to check if rate has been changed.
          -1: Rate lower
           0: No change
          +1: Rate hike
        """
        import numpy as np
        from tqdm.auto import tqdm

        # fedrates = fedrates.copy()[fedrates.index >= self.calendar.index.min()]
        fedrates["rate"] = fedrates[series_id].shift(-3)
        fedrates["prv_rate"] = fedrates[series_id].shift(1)
        fedrates["rate_change"] = fedrates["rate"] - fedrates["prv_rate"]
        fedrates["rate_decision"] = np.where(
            fedrates["rate_change"] > 0, 1, np.where(fedrates["rate_change"] < 0, -1, 0)
        )
        index_name = "index"
        columns = ["rate", "rate_change", "rate_decision"]
        fomc_calendar = self.calendar.copy()
        results = []
        for i, row in tqdm(fomc_calendar.iterrows(), total=fomc_calendar.shape[0]):
            src_ix = fedrates[fedrates.index == row.name].index.min()
            if not pd.isnull(src_ix):
                data = fedrates.loc[src_ix, columns].to_dict()
                data[index_name] = row.name
                results.append(data)
            else:
                data = {col: None for col in columns}
                data[index_name] = row.name
                results.append(data)

        results = pd.DataFrame(results)
        results.set_index(index_name, inplace=True)
        fomc_calendar[columns] = results[columns]
        fomc_calendar = fomc_calendar[fomc_calendar["rate"].notnull()]

        eKonf.save_data(fomc_calendar, self.calendar_filepath)
        self.calendar = fomc_calendar
        return fomc_calendar.copy()

    def add_unconventionals_to_calendar(self):
        """
        Add the unconventionals to the calendar.
          -2: Unconventional easing
          -1: Rate lower
           0: No change
          +1: Rate hike
          +2: Unconventional tightening
        """

        fomc_calendar = self.calendar.copy()
        unconventionals = self.unconventionals.copy().set_index("date")
        columns = ["unscheduled", "speaker", "rate", "rate_decision", "rate_change"]

        for ix in unconventionals.index:
            fomc_calendar.loc[ix, columns] = unconventionals.loc[ix, columns]

        columns = ["forecast", "confcall"]
        fomc_calendar[columns] = fomc_calendar[columns].fillna(False)
        fomc_calendar.sort_index(ascending=True, inplace=True)
        fomc_calendar["rate_changed"] = fomc_calendar["rate_decision"].apply(
            lambda x: 0 if x == 0 else 1
        )

        eKonf.save_data(fomc_calendar, self.calendar_filepath)
        self.calendar = fomc_calendar
        return fomc_calendar.copy()

    @staticmethod
    def add_available_latest(target, source, name, columns, date_offset, window=1):
        from dateutil.relativedelta import relativedelta

        date_offset = relativedelta(**date_offset)
        index_name = target.index.name
        source_ma = source.rolling(window).mean()

        results = []
        for i, row_tgt in tqdm(target.iterrows(), total=target.shape[0]):
            src_ix = source_ma[source_ma.index < row_tgt.name - date_offset].index.max()
            if not pd.isnull(src_ix):
                data = source_ma.loc[src_ix, columns].to_dict()
                data[name + "_date"] = source_ma.loc[src_ix].name
                data[index_name] = row_tgt.name
                results.append(data)
            else:
                data = {col: None for col in columns}
                data[name + "_date"] = None
                data[index_name] = row_tgt.name
                results.append(data)
        if target.shape[0] != len(results):
            print(
                "target has {} rows but returned {} rows from source!".format(
                    target.shape[0], len(results)
                )
            )
        results = pd.DataFrame(results)
        results.set_index(index_name, inplace=True)
        return target.merge(results, how="left", left_index=True, right_index=True)

    def find_speaker_of_sections(self, data):
        """
        Find the speaker of each section.
        """

        def find_speaker(row):
            content_type = row.content_type
            speaker = row.speaker
            text = row.text
            if content_type in ["fomc_press_conf", "fomc_meeting_script"]:
                match = re.findall(r"(^[A-Za-zŞ. ]*[A-Z]{3}).\d? (.*)", text)
                if len(match) == 0:
                    match = re.findall(r"(^[A-Za-zŞ. ]*[A-Z]{3}).\d(.*)", text)
                if len(match) == 1:
                    speaker, text = match[0]
                    return speaker
                return None
            else:
                return speaker

        df = data.copy()
        df["speaker"] = df.apply(find_speaker, axis=1)
        return df

    def postprocess_metadata(self, data):
        """
        - Add type
        - Add rate, decision (for meeting documents, None for the others)
        - Add next meeting date, rate and decision
        """

        def is_meeting_doc(content_type):
            if content_type in (
                "fomc_statement",
                "fomc_minutes",
                "fomc_press_conf",
                "fomc_meeting_script",
            ):
                return True
            elif content_type in ("fomc_speech", "fomc_testimony", "fomc_beigebook"):
                return False
            else:
                log.warning(f"Invalid doc_type [{content_type}] is given!")
                return None

        df = data.copy()

        df["decision"] = df.apply(
            lambda x: self._get_rate_change(x["date"])
            if is_meeting_doc(x["content_type"])
            else None,
            axis=1,
        )
        df["rate"] = df.apply(
            lambda x: self._get_rate(x["date"])
            if is_meeting_doc(x["content_type"])
            else None,
            axis=1,
        )
        df["recent_meeting"] = df["date"].map(
            lambda x: self._get_recent_meeting_date(x)
        )
        df["recent_decision"] = df["recent_meeting"].map(
            lambda x: self._get_rate_change(x)
        )
        df["recent_rate"] = df["recent_meeting"].map(lambda x: self._get_rate(x))
        df["next_meeting"] = df["date"].map(lambda x: self._get_next_meeting_date(x))
        df["next_decision"] = df["next_meeting"].map(lambda x: self._get_rate_change(x))
        df["next_rate"] = df["next_meeting"].map(lambda x: self._get_rate(x))

        return df

    def _get_rate_change(self, article_date):
        """
        Returns rate change decision of the FOMC Decision for the given date x.
        x should be of datetime type or yyyy-mm-dd format string.
        """
        if isinstance(article_date, str):
            article_date = datetime.strptime(article_date, "%Y-%m-%d")

        if article_date in self.calendar.index:
            return self.calendar.loc[article_date]["rate_decision"]
        else:
            return None

    def _get_rate(self, article_date):
        """
        Returns rate of the FOMC Decision for the given date x.
        x should be of datetime type or yyyy-mm-dd format string.
        """
        if isinstance(article_date, str):
            article_date = datetime.strptime(article_date, "%Y-%m-%d")

        if article_date in self.calendar.index:
            return self.calendar.loc[article_date]["rate"]
        else:
            return None

    def _get_next_meeting_date(self, article_date):
        """
        Returns the next fomc meeting date for the given date x, referring to fomc_calendar DataFrame.
        Usually FOMC Meetings takes two days, so it starts searching from x+2.
        x should be of datetime type or yyyy-mm-dd format string.
        """
        import datetime as dt

        if isinstance(article_date, str):
            article_date = datetime.strptime(article_date, "%Y-%m-%d")

        # Add two days to get the day after next
        article_date += dt.timedelta(days=2)

        cal = self.calendar.copy()
        if cal.index.min() > article_date:
            # If the date is older than the first FOMC Meeting, do not return any date.
            return None
        else:
            dt_ix = cal[cal.index > article_date].index.min()
            if not pd.isnull(dt_ix):
                return dt_ix
            else:
                return None

    def _get_recent_meeting_date(self, article_date):
        """
        Returns the most recent fomc meeting date for the given date x, referring to fomc_calendar DataFrame.
        Usually FOMC Meetings takes two days, so it starts searching from x+2.
        x should be of datetime type or yyyy-mm-dd format string.
        """
        import datetime as dt

        if isinstance(article_date, str):
            article_date = datetime.strptime(article_date, "%Y-%m-%d")

        # Add two days to get the day after next
        article_date += dt.timedelta(days=2)

        cal = self.calendar.copy()
        if cal.index.min() > article_date:
            # If the date is older than the first FOMC Meeting, do not return any date.
            return None
        else:
            dt_ix = cal[cal.index < article_date].index.max()
            if not pd.isnull(dt_ix):
                return dt_ix
            else:
                return None

    @staticmethod
    def get_irf(data, nlags=4, nirf=20):
        import statsmodels.api as sm

        # Dimensions
        nobs = data.shape[0] - nlags
        neqs = data.shape[1]

        # Estimate via statsmodels VAR
        mod = sm.tsa.VAR(data)
        res = mod.fit(maxlags=nlags, ic=None, trend="n")

        # Variance / Covariance matrix
        omega = res.sigma_u

        # Take LDL decomposition of the variance / covariance matrix
        L = np.linalg.cholesky(omega)
        D = np.diag(L.diagonal() ** 2)
        L[np.diag_indices(L.shape[0])] = 1

        # Storage
        irf_tilde = np.zeros((nirf + nlags + 1, neqs, neqs))
        irf = np.zeros((nirf, neqs, neqs))

        # Initial values
        irf_tilde[nlags] = np.eye(neqs)
        irf[0] = L

        # Iteratively create IRFs
        for i in range(nirf):
            for j in range(nlags):
                irf_tilde[i + nlags + 1] += np.dot(
                    res.params.iloc[j * neqs : (j + 1) * neqs, :].T,
                    irf_tilde[i + nlags - j],
                )
            irf[i] = np.dot(irf_tilde[i + nlags], L)

        irfs = {}
        for col in data.columns:
            ix = data.columns.get_loc(col)
            irf_df = pd.DataFrame(-irf[:, :, ix])
            irf_df.columns = data.columns
            irfs[col] = irf_df

        return irfs

    @staticmethod
    def plot_irf(irfs, impulse_name, ncols=2, figsize=(16, 12), title=None):
        irf = irfs[impulse_name]

        title = title or f"Impulse Response of {impulse_name} shock"
        names = irf.columns.tolist()
        neqs = len(names)

        cfg = eKonf.compose("visualize/plot=lineplot")

        cfg.figure.figsize = figsize
        cfg.subplots.ncols = ncols
        cfg.subplots.nrows = int(np.ceil(neqs / ncols))
        lineplot = cfg.lineplot.copy()
        ax = cfg.ax.copy()
        cfg.plots = []
        cfg.axes = []

        for i, name in enumerate(names):
            plot = lineplot.copy()
            plot.y = name
            plot.axno = i
            cfg.plots.append(plot)
            _ax = ax.copy()
            _ax.title = name
            _ax.axno = i
            cfg.axes.append(_ax)
        cfg.figure.super.title = title

        eKonf.instantiate(cfg, data=irf)
